/* $Id: action.c,v 1.190 2010/03/12 14:27:37 oohara Exp $ */

#include <stdio.h>
/* free */
#include <stdlib.h>

#include "world.h"
#include "creature.h"
#include "dungeon.h"
#include "save.h"
#include "creature-detail.h"
#include "scroll-detail.h"
#include "enchantment-detail.h"
#include "grid-detail.h"

#include "action.h"

static int player_pickup(world *wp);
static int player_know_scroll(world *wp, int type);
static int player_know_inventory(world *wp);
static int detonation(world *wp, int z, int x, int y, int attacker_id);
static int map_floor(world *wp, int z);
static int attack_creature_here(world *wp, int attacker_id,
                                int z, int x, int y, int melee);
static int melee_attack(world *wp, int attacker_id, int dx, int dy);
static int shadow_attack_hits_here(world *wp, int x, int y);
static int missile_attack(world *wp, int attacker_id, int dx, int dy);
static int scroll_effect(world *wp, int type, int dx, int dy);
static int order_shadow(world *wp, int dx, int dy);
static int action_exec_walk(world *wp, creature *who);
static int action_exec_use_scroll(world *wp, creature *who);
static int action_exec_save(world *wp, creature *who);
static int action_exec_climb_stair(world *wp, creature *who);
static int action_exec_create_shadow(world *wp, creature *who);
static int action_exec_destroy_shadow(world *wp, creature *who);

static int
player_pickup(world *wp)
{
  int i;
  char buf[128];
  struct creature_detail crd_player;
  creature *player = NULL;
  char *sname = NULL;

  if (wp == NULL)
    return 1;

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];
  creature_detail_get(player->type, &crd_player);

  for (i = 0; i < wp->obj_size; i++)
  {
    if (object_is_dead(wp, i))
      continue;
    if ((wp->obj[i]->z != player->z)
        || (wp->obj[i]->x != player->x)
        || (wp->obj[i]->y != player->y))
      continue;
    if (add_inventory(wp, wp->obj[i]->type, wp->obj[i]->quantity) < 0)
    {
      fprintf(stderr, "player_pickup: add_inventory failed\n");
      return 1;
    }
    sname = scroll_name(wp, wp->obj[i]->type);
    sprintf(buf, "%s picks up %d scroll of %s",
            crd_player.name, wp->obj[i]->quantity, sname);
    add_log(wp, buf);
    if (sname != NULL)
    {
      free(sname);
      sname = NULL;
    }
    wp->obj[i]->quantity = 0;
  }

  return 0;
}

static int
player_know_scroll(world *wp, int type)
{
  struct scroll_detail scd;
  char buf[128];
  char *sname = NULL;

  if (wp == NULL)
    return 1;

  if (!scroll_is_unknown(wp, type))
    return 0;
  if (creature_is_dead(wp, wp->player_id))
    return 0;

  scroll_detail_get(type, &scd);
  sname = scroll_name(wp, type);
  sprintf(buf, "scroll of %s is scroll of %s", sname, scd.real_name);
  add_log(wp, buf);
  if (sname != NULL)
  {
    free(sname);
    sname = NULL;
  }

  wp->label[type] = -1;

  return 0;
}

static int
player_know_inventory(world *wp)
{
  int i;

  if (wp == NULL)
    return 1;

  if (creature_is_dead(wp, wp->player_id))
    return 0;

  for (i = 0; i < wp->inv_size; i++)
  {
    if (inventory_is_dead(wp, i))
      continue;
    player_know_scroll(wp, wp->inv[i]->type);
  }

  return 0;
}

/* If (z, x, y) is a wall, this effect _can_ damage the monster
 * behind the wall; this is not a rocket.
 *
 *     #
 * @...*d
 *     #
 */
static int
detonation(world *wp, int z, int x, int y, int attacker_id)
{
  int i;
  int dx;
  int dy;
  int dist;

  if (wp == NULL)
    return 1;

  for (dist = 0; dist <= 1; dist++)
  {
    for (dx = -1; dx <= 1; dx++)
    {
      for (dy = -1; dy <= 1; dy++)
      {
        if (distance(dx, dy, 0, 0) != dist)
          continue;
        if (grid_is_illegal(wp, z, x + dx, y + dy))
          continue;
        for (i = 0; i < wp->cr_size; i++)
        {
          if (creature_is_dead(wp, i))
            continue;
          if (wp->cr[i]->z != z)
            continue;
          if (wp->cr[i]->x != x + dx)
            continue;
          if (wp->cr[i]->y != y + dy)
            continue;

          creature_take_damage(wp, i, 18,
                               attacker_id, 1);
        }
      }
    }
  }

  return 0;
}

static int
map_floor(world *wp, int z)
{
  int x;
  int y;

  if (wp == NULL)
    return 1;
  if ((z < 0) || (z >= wp->size_z))
    return 1;

  if (creature_is_dead(wp, wp->player_id))
    return 0;

  for (x = 0; x < wp->size_x; x++)
  {
    for (y = 0; y < wp->size_y; y++)
    {
      wp->grid[z][x][y] |= GRID_PLAYER_SEEN;
    }
  }

  return 0;
}

/* return 1 if the attack should stop, 0 otherwise */
static int
attack_creature_here(world *wp, int attacker_id,
                     int z, int x, int y, int melee)
{
  int i;
  int dam;
  int stop;
  int reflect;
  struct creature_detail crd_attacker;
  struct creature_detail crd_target;
  char buf[128];

  if (wp == NULL)
    return 1;
  if ((attacker_id < 0) || (attacker_id >= wp->cr_size))
    return 1;
  if (wp->cr[attacker_id] == NULL)
    return 1;
  if (grid_is_illegal(wp, z, x, y))
    return 1;

  if ((wp->cr[attacker_id]->type == CR_PLAYER_SHADOW)
      && (!shadow_attack_hits_here(wp, x, y)))
    return 0;

  creature_detail_get(wp->cr[attacker_id]->type, &crd_attacker);
  if (melee)
    dam = crd_attacker.power_melee;
  else
    dam = crd_attacker.power_missile;

  if (wp->cr[attacker_id]->enchant_duration > 0)
  {
    switch (wp->cr[attacker_id]->enchant_type)
    {
    case ENC_WEAK:
      dam /= 4;
      break;
    default:
      break;
    }
  }

  if (dam <= 0)
    return 1;

  stop = 0;
  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == attacker_id)
      continue;
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->z != z)
      continue;
    if (wp->cr[i]->x != x)
      continue;
    if (wp->cr[i]->y != y)
      continue;
    if (creature_is_friendly(wp, i))
    {
      if (creature_is_friendly(wp, attacker_id))
        continue;
    }
    else
    {
      if (!creature_is_friendly(wp, attacker_id))
        continue;
    }

    if (wp->cr[i]->type == CR_DISENCHANTER)
    {
      if ((wp->cr[attacker_id]->enchant_duration > 0)
          && (wp->cr[attacker_id]->enchant_type > 0))
        continue;
    }

    reflect = 0;
    creature_detail_get(wp->cr[i]->type, &crd_target);

    switch (wp->cr[i]->type)
    {
    case CR_CRYSTAL_FLOWER:
      if (!melee)
        reflect = 1;
      break;
    default:
      break;
    }

    if (wp->cr[i]->enchant_duration > 0)
    {
      switch (wp->cr[i]->enchant_type)
      {
      case ENC_REFLECT:
        if (!melee)
          reflect = 1;
          break;
      default:
        break;
      }
    }

    if (!melee)
      stop = 1;
    if (reflect)
    {
      sprintf(buf, "%s reflects %s attack of %s",
              crd_target.name,
              (melee) ? "melee" : "missile",
              crd_attacker.name);
      add_log(wp, buf);
      creature_take_damage(wp, attacker_id, dam, i, 0);
      /* no side effects */
      continue;
    }
    else
    {  
      sprintf(buf, "%s %s %s",
              crd_attacker.name,
              (melee) ? "hits" : "shoots",
              crd_target.name);
      add_log(wp, buf);
      creature_take_damage(wp, i, dam, attacker_id, 0);
    }

    switch (wp->cr[attacker_id]->type)
    {
    case CR_DISENCHANTER:
      creature_add_enchantment(wp, i,
                               ENC_ANTIMAGIC, 2);
      break;
    case CR_JELLY:
      if (melee)
        creature_add_enchantment(wp, i,
                                 ENC_POISON, 2);
      break;
    case CR_IRON_GOLEM:
      if (melee)
        creature_add_enchantment(wp, i,
                                 ENC_STUN, 4);
      break;

    default:
      break;
    }
  }

  return stop;
}

static int
melee_attack(world *wp, int attacker_id, int dx, int dy)
{
  int z;
  int x;
  int y;

  if (wp == NULL)
    return 1;
  if ((dx < -1) || (dx > 1))
    return 1;
  if ((dy < -1) || (dy > 1))
    return 1;
  if ((dx == 0) && (dy == 0))
    return 1;

  if (creature_is_dead(wp, attacker_id))
    return 0;

  if ((wp->cr[attacker_id]->enchant_duration > 0))
  {
    switch (wp->cr[attacker_id]->enchant_type)
    {
    case ENC_FEAR:
      return 0;
      break;
    default:
      break;
    }
  }

  z = wp->cr[attacker_id]->z;
  x = wp->cr[attacker_id]->x;
  y = wp->cr[attacker_id]->y;

  attack_creature_here(wp, attacker_id, z, x + dx, y + dy, 1);
  if (dx == 0)
  {
    attack_creature_here(wp, attacker_id, z, x + 1, y + dy, 1);
    attack_creature_here(wp, attacker_id, z, x - 1, y + dy, 1);
  }
  else if (dy == 0)
  {
    attack_creature_here(wp, attacker_id, z, x + dx, y + 1, 1);
    attack_creature_here(wp, attacker_id, z, x + dx, y + -1, 1);
  }
  else
  {
    attack_creature_here(wp, attacker_id, z, x + dx, y, 1);
    attack_creature_here(wp, attacker_id, z, x, y + dy, 1);
  }

  return 0;
}

/* assumes that all shadows are in your current floor */
static int
shadow_attack_hits_here(world *wp, int x, int y)
{
  if (wp == NULL)
    return 0;

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  if (distance(x, y, wp->cr[wp->player_id]->x, wp->cr[wp->player_id]->y) > 8)
    return 0;

  return 1;
}

static int
missile_attack(world *wp, int attacker_id, int dx, int dy)
{
  int i;
  int z;
  int x;
  int y;

  if (wp == NULL)
    return 1;
  if ((dx < -1) || (dx > 1))
    return 1;
  if ((dy < -1) || (dy > 1))
    return 1;
  if ((dx == 0) && (dy == 0))
    return 1;

  if (creature_is_dead(wp, attacker_id))
    return 0;

  z = wp->cr[attacker_id]->z;
  x = wp->cr[attacker_id]->x;
  y = wp->cr[attacker_id]->y;

  for (i = 1; i <= 6; i++)
  {
    if ((wp->cr[attacker_id]->enchant_duration > 0))
    {
      if (wp->cr[attacker_id]->enchant_type == ENC_FEAR)
      {
        if (i == 1)
          continue;
      }
    }
    if (attack_creature_here(wp, attacker_id,
                             z, x + dx * i, y + dy * i, 0) != 0)
      break;
    if (grid_blocks_beam(wp, z, x + dx * i, y + dy * i))
      break;
  }

  return 0;
}

/* assumes that only the player can use a scroll */
static int
scroll_effect(world *wp, int type, int dx, int dy)
{
  int i;
  int j;
  int z;
  int x;
  int y;
  int no_more_creature;
  int no_more_grid;
  creature *who = NULL;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
    return 1;
  if (wp->cr[wp->player_id] == NULL)
    return 1;
  player = wp->cr[wp->player_id];

  if ((dx == 0) && (dy == 0))
  {
    switch (type)
    {
    case SC_KNOWLEDGE:
      player_know_inventory(wp);
      return 0;
      break;
    default:
      break;
    }
  }

  no_more_creature = 0;
  no_more_grid = 0;
  for (i = 1; i <= 6; i++)
  {
    z = player->z;
    x = player->x + dx * i;
    y = player->y + dy * i;

    if (grid_is_illegal(wp, z, x, y))
      break;

    for (j = 0; j < wp->cr_size; j++)
    {
      if (creature_is_dead(wp, j))
        continue;
      who = wp->cr[j];
      if (who->z != z)
        continue;
      if (who->x != x)
        continue;
      if (who->y != y)
        continue;
      switch (type)
      {
      case SC_TELEPORT:
        creature_teleport(wp, who->id);
        no_more_grid = 1;
        break;
      case SC_POISON:
        creature_add_enchantment(wp, who->id,
                                 ENC_POISON, 4);
        no_more_grid = 1;
        break;
      case SC_SPEED:
        creature_add_enchantment(wp, who->id,
                                 ENC_FAST, 11);
        no_more_grid = 1;
        break;
      case SC_SLOWNESS:
        creature_add_enchantment(wp, who->id,
                                 ENC_SLOW, 10);
        no_more_grid = 0;
        break;
      case SC_REGENERATION:
        creature_add_enchantment(wp, who->id,
                                 ENC_REGENERATE, 33);
        no_more_grid = 1;
        break;
      case SC_MIGHT:
        creature_add_enchantment(wp, who->id,
                                 ENC_MIGHT, 13);
        no_more_grid = 1;
        break;
      case SC_WEAKNESS:
        creature_add_enchantment(wp, who->id,
                                 ENC_WEAK, 12);
        no_more_grid = 0;
        break;
      case SC_REFLECTION:
        creature_add_enchantment(wp, who->id,
                                 ENC_REFLECT, 7);
        no_more_grid = 1;
        break;
      case SC_INVULNERABILITY:
        creature_add_enchantment(wp, who->id,
                                 ENC_INVULNERABLE, 5);
        no_more_grid = 1;
        break;
      case SC_ANTIMAGIC:
        creature_add_enchantment(wp, who->id,
                                 ENC_ANTIMAGIC, 9);
        no_more_grid = 1;
        break;
      case SC_STUNNING:
        creature_add_enchantment(wp, who->id,
                                 ENC_STUN, 4);
        no_more_grid = 1;
        break;
      case SC_FEAR:
        creature_add_enchantment(wp, who->id,
                                 ENC_FEAR, 16);
        no_more_grid = 1;
        break;
      case SC_LIGHTNING:
        creature_take_damage(wp, who->id, 12,
                             player->id, 1);
        no_more_grid = 0;
        break;
      case SC_DETONATION:
        detonation(wp, z, x, y, player->id);
        no_more_creature = 1;
        no_more_grid = 1;
        break;
      case SC_KNOWLEDGE:
        map_floor(wp, z);
        no_more_creature = 1;
        no_more_grid = 1;
        break;
      default:
        break;
      }
      if (no_more_creature)
        break;
    }

    if ((dx == 0) && (dy == 0))
      break;
    if (grid_blocks_beam(wp, z, x, y))
    {
      switch (type)
      {
      case SC_DETONATION:
        detonation(wp, z, x, y, player->id);
        break;
      case SC_KNOWLEDGE:
        map_floor(wp, z);
        break;
      default:
        break;
      }
      no_more_grid = 1;
    }
    if ((no_more_creature) || (no_more_grid))
      break;
  }

  return 0;
}

/* assumes only the player can use shadow */
static int
order_shadow(world *wp, int dx, int dy)
{
  int i;

  if (wp == NULL)
    return 1;
  if ((dx < -1) || (dx > 1))
    return 1;
  if ((dy < -1) || (dy > 1))
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->type != CR_PLAYER_SHADOW)
      continue;
    switch (wp->shadow_mode)
    {
    case 0:  
      wp->cr[i]->act[0] = ACTION_WALK;
      wp->cr[i]->act[1] = dx;
      wp->cr[i]->act[2] = dy;
      break;
    case 1:  
      wp->cr[i]->act[0] = ACTION_WALK;
      wp->cr[i]->act[1] = -dx;
      wp->cr[i]->act[2] = -dy;
      break;
    default:
      break;
    }
  }

  return 0;
}

/* return 1 if the action is legal, 0 otherwise */
int
action_check(world *wp, int creature_id, int *act)
{
  int x;
  int y;
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "action_check: wp is NULL\n");
    return 0;
  }
  if (act == NULL)
  {
    fprintf(stderr, "action_check: act is NULL\n");
    return 0;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "action_check: strange creature_id (%d)\n",
            creature_id);
    return 0;
  }

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];

  if (act[0] == 0)
    return 0;

  switch(act[0])
  {
  case ACTION_STAY:
    return 1;
    break;
  case ACTION_WALK:
    /* act
     * [1]: dx
     * [2]: dy
     */
    if ((act[1] == 0) && (act[2] == 0))
      return 1;
    if ((act[1] < -1) || (act[1] > 1))
      return 0;
    if ((act[2] < -1) || (act[2] > 1))
      return 0;
    x = who->x + act[1];
    y = who->y + act[2];
    if (grid_is_illegal(wp, who->z, x, y))
      return 0;
    if (grid_blocks_creature(wp, who->z, x, y))
      return 0;
    return 1;
    break;
  case ACTION_USE_SCROLL:
    /* act
     * [1] inventory chosen
     * [2] dx
     * [3] dy
     */
    if (who->id != wp->player_id)
      return 0;
    if (inventory_is_dead(wp, act[1]))
      return 0;
    if ((act[2] < -1) || (act[2] > 1))
      return 0;
    if ((act[3] < -1) || (act[3] > 1))
      return 0;
    return 1;
    break;
  case ACTION_SAVE:
    if (who->id == wp->player_id)
      return 1;
    return 0;
    break;
  case ACTION_CLIMB_STAIR:
    if (who->id != wp->player_id)
      return 0;
    if (!find_stair_destination(wp,
                                who->z, who->x, who->y,
                                NULL, NULL, NULL))
      return 0;
    return 1;
    break;
  case ACTION_CREATE_SHADOW:
    /* act
     * [1] dx
     * [2] dy
     */
    if (who->id != wp->player_id)
      return 0;
    if (creature_count_this_type(wp, CR_PLAYER_SHADOW) >= 2)
      return 0;
    if ((act[1] == 0) && (act[2] == 0))
      return 0;
    if (grid_blocks_creature(wp,
                             who->z, who->x + act[1], who->y + act[2]))
      return 0;
    if (grid_has_creature(wp,
                          who->z, who->x + act[1], who->y + act[2]))
      return 0;
    return 1;
    break;
  case ACTION_DESTROY_SHADOW:
    if (who->id != wp->player_id)
      return 0;
    if (creature_count_this_type(wp, CR_PLAYER_SHADOW) <= 0)
      return 0;
    return 1;
  default:
    return 0;
    break;
  }

  fprintf(stderr, "action_check: should not reach here\n");
  return 0;
}

/*
 * these action_exec_* functions assume that the action is legal
 *
 */
static int
action_exec_walk(world *wp, creature *who)
{
  int x;
  int y;
  int occupied;
  int do_melee_attack;
  int do_missile_attack;

  if (wp == NULL)
    return 1;
  if (who == NULL)
    return 1;

  if ((who->act[1] == 0) && (who->act[2] == 0))
    return 0;

  do_melee_attack = 0;
  do_missile_attack = 0;
  x = who->x + who->act[1];
  y = who->y + who->act[2];

  occupied = 0;
  if (grid_blocks_creature(wp, who->z, x, y))
    occupied = 1;
  if (grid_has_creature(wp, who->z, x, y))
    occupied = 1;

  switch (who->type)
  {
  case CR_MOLD:
  case CR_TURRET:
  case CR_CRYSTAL_FLOWER:
    occupied = 1;
    break;
  default:
    break;
  }

  if (who->enchant_duration > 0)
  {
    switch (who->enchant_type)
    {
    case ENC_STUN:
      occupied = 1;
      break;
    default:
      break;
    }
  }

  if (!occupied)
  {
    creature_change_xy(wp, who->id, x, y);

    do_missile_attack = 1;
  }
  else
  {
    do_melee_attack = 1;
  }

  if ((do_melee_attack) || (do_missile_attack))
  {
    switch (who->type)
    {
    case CR_WARLORD:
      do_melee_attack = 1;
      do_missile_attack = 1;
      break;
    case CR_TURRET:
      do_melee_attack = 0;
      do_missile_attack = 1;
      break;
    default:
      break;
    }

    if ((who->enchant_duration > 0)
        && (who->enchant_type == ENC_MIGHT))
    {
      do_melee_attack = 1;
      do_missile_attack = 1;
    }
  }

  if (do_melee_attack)
    melee_attack(wp, who->id, who->act[1], who->act[2]);
  if (do_missile_attack)
    missile_attack(wp, who->id, who->act[1], who->act[2]);

  if (who->id == wp->player_id)
    order_shadow(wp, who->act[1], who->act[2]);

  return 0;
}

static int
action_exec_use_scroll(world *wp, creature *who)
{
  struct creature_detail crd;
  char buf[128];
  char *sname = NULL;

  if (wp == NULL)
    return 1;
  if (who == NULL)
    return 1;

  creature_detail_get(who->type, &crd);
  sname = scroll_name(wp, wp->inv[who->act[1]]->type);
  sprintf(buf, "%s uses scroll of %s", crd.name, sname);
  add_log(wp, buf);
  if (sname != NULL)
  {
    free(sname);
    sname = NULL;
  }
  player_know_scroll(wp, wp->inv[who->act[1]]->type);
  scroll_effect(wp, wp->inv[who->act[1]]->type, who->act[2], who->act[3]);
  (wp->inv[who->act[1]]->quantity)--;

  return 0;
}

static int
action_exec_save(world *wp, creature *who)
{
  if (wp == NULL)
    return 1;
  if (who == NULL)
    return 1;

  if (save_game(wp) == 0)
  {
    add_log(wp, "game saved successfully");
    wp->should_quit = 1;
  }
  else
  {
    add_log(wp, "game save FAILED");
  }

  return 0;
}

static int
action_exec_climb_stair(world *wp, creature *who)
{
  int i;
  int dest_z;
  int dest_x;
  int dest_y;
  struct creature_detail crd_you;
  struct creature_detail crd_target;
  char buf[128];

  if (wp == NULL)
    return 1;
  if (who == NULL)
    return 1;

  if (!find_stair_destination(wp,
                              who->z, who->x, who->y,
                              &dest_z, &dest_x, &dest_y))
    return 1;

  creature_detail_get(who->type, &crd_you);
  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == who->id)
      continue;
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->z != dest_z)
      continue;
    if (wp->cr[i]->x != dest_x)
      continue;
    if (wp->cr[i]->y != dest_y)
      continue;

    creature_detail_get(wp->cr[i]->type, &crd_target);
    sprintf(buf, "%s pushes past %s", crd_you.name, crd_target.name);
    add_log(wp, buf);
    creature_change_z(wp, i, who->z);
    wp->cr[i]->x = who->x;
    wp->cr[i]->y = who->y;
  }

  sprintf(buf, "%s enters B%dF", crd_you.name, dest_z);
  add_log(wp, buf);

  creature_change_z(wp, who->id, dest_z);
  creature_change_xy(wp, who->id, dest_x, dest_y);

  return 0;
}

static int
action_exec_create_shadow(world *wp, creature *who)
{
  int n;
  char buf[128];
  struct creature_detail crd;
  struct creature_detail crd_shadow;

  if (wp == NULL)
    return 1;
  if (who == NULL)
    return 1;

  n = add_new_creature(wp, CR_PLAYER_SHADOW,
                       who->z, who->x + who->act[1], who->y + who->act[2],
                       who->wait);
  if (n < 0)
  {
    fprintf(stderr, "action_exec_cast_shadow: add_new_creature failed\n");
    return 1;
  }

  creature_detail_get(who->type, &crd);
  creature_detail_get(wp->cr[n]->type, &crd_shadow);
  sprintf(buf, "%s creates %s", crd.name, crd_shadow.name);
  add_log(wp, buf);

  return 0;
}

static int
action_exec_destroy_shadow(world *wp, creature *who)
{
  int i;
  char buf[128];
  struct creature_detail crd;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->type != CR_PLAYER_SHADOW)
      continue;

    wp->cr[i]->hp = 0;
  }

  creature_detail_get(who->type, &crd);
  sprintf(buf, "%s destroys shadows", crd.name);
  add_log(wp, buf);

  return 0;
}

/* return the amount of time taken by the action */
int
action_exec(world *wp, int creature_id)
{
  int ut;
  int ut_taken;
  int i;
  struct creature_detail crd;
  char buf[128];
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "action_exec: wp is NULL\n");
    return 0;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "action_exec: strange creature_id (%d)\n",
            creature_id);
    return 0;
  }

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];
  creature_detail_get(who->type, &crd);

  if (who->act[0] == 0)
    return 0;

  ut = 4;
  ut_taken = 0;

  if ((who->id == wp->player_id)
      && (candle_is_in_the_last(wp)))
    ut = (ut + 1) / 2;

  switch (who->type)
  {
  case CR_PLAYER_SHADOW:
  if ((who->id != wp->player_id)
      && (candle_is_in_the_last(wp)))
    ut = (ut + 1) / 2;
    break;
  case CR_IRON_GOLEM:
    ut *= 2;
    break;
  default:
    break;
  }

  if (who->enchant_duration > 0)
  {
    switch (who->enchant_type)
    {
    case ENC_FAST:
      ut = (ut + 1) / 2;
      break;
    case ENC_SLOW:
      ut *= 2;
      break;
    default:
      break;
    }
  }

  if (ut < 1)
    ut = 1;

  if (!action_check(wp, who->id, who->act))
  {
    sprintf(buf, "%s tries something illegal", crd.name);
    add_log(wp, buf);
    return ut;
  }

  switch(who->act[0])
  {
  case ACTION_STAY:
    ut_taken = ut;
    break;
  case ACTION_WALK:
    ut_taken = ut * 2;
    action_exec_walk(wp, who);
    break;
  case ACTION_USE_SCROLL:
    ut_taken = ut * 2;
    action_exec_use_scroll(wp, who);
    break;
  case ACTION_SAVE:
    ut_taken = 0;
    action_exec_save(wp, who);
    break;
  case ACTION_CLIMB_STAIR:
    ut_taken = ut * 2;
    action_exec_climb_stair(wp, who);
    break;
  case ACTION_CREATE_SHADOW:
    ut_taken = ut;
    action_exec_create_shadow(wp, who);
    break;
  case ACTION_DESTROY_SHADOW:
    ut_taken = ut;
    action_exec_destroy_shadow(wp, who);
    break;
  default:
    sprintf(buf, "%s tries something undefined", crd.name);
    add_log(wp, buf);
    ut_taken = ut;
    break;
  }

  if (who->id == wp->player_id)
  {
    player_pickup(wp);
  }

  for (i = 0; i < 4; i++)
    who->act[i] = 0;

  return ut_taken;
}
