/* $Id: build.c,v 1.127 2010/03/12 16:01:29 oohara Exp $ */

#include <stdio.h>
/* rand */
#include <stdlib.h>

#include "world.h"
#include "dungeon.h"
#include "grid-detail.h"
#include "scroll-detail.h"
#include "creature-detail.h"

#include "build.h"

static int randomize_int_array(int *array, int size);
static int fill_floor(world *wp, int z, int ax, int ay, int bx, int by,
                      int type);
static int build_grid_bsp_recursive(world *wp, int z,
                                    int ax, int ay,
                                    int bx, int by,
                                    int recursion_depth,
                                    int recursion_depth_max);
static int build_grid_bsp(world *wp, int z);
static int build_grid_haven(world *wp, int z);
static int build_grid_rubble(world *wp, int z);
static int build_grid_glass_castle(world *wp, int z);

static int
randomize_int_array(int *array, int size)
{
  int i;
  int n;
  int temp;

  if (array == NULL)
    return 1;
  if (size <= 0)
    return 1;

  for (i = 0; i < size; i++)
  {
    n = rand() % (size - i);
    temp = array[i + n];
    array[i + n] = array[i];
    array[i] = temp;
  }

  return 0;
}

static int
fill_floor(world *wp, int z, int ax, int ay, int bx, int by,
           int type)
{
  int x;
  int y;

  if (wp == NULL)
    return 1;
  if (grid_is_illegal(wp, z, ax, ay))
    return 1;
  if (grid_is_illegal(wp, z, bx, by))
    return 1;

  for (x = ax; x <= bx; x++)
  {
    for (y = ay; y <= by; y++)
    {
      wp->grid[z][x][y] = type;
    }
  }

  return 0;
}

static int
build_grid_bsp_recursive(world *wp, int z,
                         int ax, int ay,
                         int bx, int by,
                         int recursion_depth,
                         int recursion_depth_max)
{
  int i;
  int x;
  int y;
  int dx;
  int dy;
  int mode;
  int ax_next[2];
  int ay_next[2];
  int bx_next[2];
  int by_next[2];
  int ax_open;
  int ay_open;
  int bx_open;
  int by_open;

  if (wp == NULL)
    return 1;
  if (grid_is_illegal(wp, z, ax, ay))
    return 1;
  if (grid_is_illegal(wp, z, bx, by))
    return 1;

  dx = bx - ax;
  dy = by - ay;
  if ((dx <= 0) || (dy <= 0))
    return 1;

  if ((recursion_depth >= recursion_depth_max)
      || (dx <= 2)
      || (dy <= 2)
      || ((dx <= 5) && (dy <= 5)))
  {
    fill_floor(wp, z, ax, ay, bx, by, GR_FLOOR);
    return 0;
  }  

  mode = recursion_depth % 2;

  if (mode == 0)
  {
    ax_next[0] = ax;
    ay_next[0] = ay;
    bx_next[0] = bx;
    by_next[0] = ay + dy / 2 - 1;
    if (dy >= 8)
      by_next[0] -= rand() % (dy / 8);

    ax_next[1] = ax;
    ay_next[1] = ay + dy / 2 + 1;
    if (dy >= 8)
      ay_next[1] += rand() % (dy / 8);
    bx_next[1] = bx;
    by_next[1] = by;
  }
  else
  {
    ax_next[0] = ax;
    ay_next[0] = ay;
    bx_next[0] = ax + dx / 2 - 1;
    if (dx >= 8)
      bx_next[0] -= rand() % (dx / 8);
    by_next[0] = by;

    ax_next[1] = ax + dx / 2 + 1;
    if (dx >= 8)
      ax_next[1] += rand() % (dx / 8);
    ay_next[1] = ay;
    bx_next[1] = bx;
    by_next[1] = by;
  }

  for (i = 0; i < 2; i++)
  {
    build_grid_bsp_recursive(wp, z,
                             ax_next[i], ay_next[i],
                             bx_next[i], by_next[i],
                             recursion_depth + 1,
                             recursion_depth_max);
  }

  do
  {
    if (mode == 0)
    {
      dx = 0;
      dy = 1;
      ax_open = ax + 1 + (rand() % (dx - 2));
      ay_open = ay_next[0] + (rand() % (by_next[0] - ay_next[0]));
      bx_open = ax_open;
      by_open = ay_next[1] + (rand() % (by_next[1] - ay_next[1]));
    }
    else
    {
      dx = 1;
      dy = 0;
      ay_open = ay + 1 + (rand() % (dy - 2));
      ax_open = ax_next[0] + (rand() % (bx_next[0] - ax_next[0]));
      by_open = ay_open;
      bx_open = ax_next[1] + (rand() % (bx_next[1] - ax_next[1]));
    }
  } while ((grid_blocks_creature(wp, z, ax_open, ay_open))
           || (grid_blocks_creature(wp, z, bx_open, by_open)));
  
  for (i = 0; 1 == 1; i++)
  {
    x = ax_open + dx * i;
    y = ay_open + dy * i;
    if ((grid_is_illegal(wp, z, x, y)))
      break;
    if ((x > bx_open) || (y > by_open))
      break;

    if (grid_blocks_creature(wp, z, x, y))
      wp->grid[z][x][y] = GR_FLOOR;
  }

  return 0;
}

static int
build_grid_bsp(world *wp, int z)
{
  int i;
  int j;
  int n;
  int x;
  int y;
  int dx;
  int dy;
  int ax;
  int ay;
  
  if (wp == NULL)
  {
    fprintf(stderr, "build_grid_bsp: wp is NULL\n");
    return 1;
  }
  if ((z < 0) || (z >= wp->size_z))
  {
    fprintf(stderr, "build_grid_bsp: strange z (%d)\n", z);
    return 1;
  }
  if (wp->size_x <= 2)
  {
    fprintf(stderr, "build_grid_bsp: wp->size_x is too small (%d)\n",
            wp->size_x);
    return 1;
  }
  if (wp->size_y <= 2)
  {
    fprintf(stderr, "build_grid_bsp: wp->size_y is too small (%d)\n",
            wp->size_y);
    return 1;
  }

  fill_floor(wp, z, 0, 0, wp->size_x - 1, wp->size_y - 1, GR_WALL);

  build_grid_bsp_recursive(wp, z,
                           1, 1,
                           wp->size_x - 2, wp->size_y - 2,
                           0, 5);

  for (i = 0; i < 32; i++)
  {
    ax = 1 + (rand() % (wp->size_x - 2));
    ay = 1 + (rand() % (wp->size_y - 2));
    if (grid_blocks_creature(wp, z, ax, ay))
      continue;

    switch (i % 4)
    {
    case 0:
      dx = 1;
      dy = 0;
      break;
    case 1:
      dx = 0;
      dy = 1;
      break;
    case 2:
      dx = -1;
      dy = 0;
      break;
    default:
      dx = 0;
      dy = -1;
      break;
    }

    n = wp->size_x;
    if (n > wp->size_y)
      n = wp->size_y;
    n -= 2;
    n = (n * 3) / 4;
    if (n < 1)
      n = 1;
    n = 1 + (rand() % n);

    for (j = 0; j < n; j++)
    {
      x = ax + dx * j;
      y = ay + dy * j;
      if ((x <= 1) || (x >= wp->size_x - 2))
        break;
      if ((y <= 1) || (y >= wp->size_y - 2))
        break;
      if (grid_blocks_creature(wp, z, x, y))
        wp->grid[z][x][y] = GR_FLOOR;
    }
  }

  for (i = 0; i < 4; i++)
  {
    do
    {
      x = 1 + (rand() % (wp->size_x - 2));
      y = 1 + (rand() % (wp->size_y - 2));
    } while (wp->grid[z][x][y] != GR_FLOOR);

    switch (i)
    {
    case 0:
      if (z == wp->size_z - 1)
        wp->grid[z][x][y] = GR_GATE_OF_DOOM;
      else if (z <= wp->size_z - 2)
        wp->grid[z][x][y] = GR_STAIR_DOWN_0;
      break;
    case 1:
      if (z <= wp->size_z - 2)
        wp->grid[z][x][y] = GR_STAIR_DOWN_1;
      break;
    case 2:
      if (z >= 1)
        wp->grid[z][x][y] = GR_STAIR_UP_0;
      break;
    default:
      if (z >= 1)
        wp->grid[z][x][y] = GR_STAIR_UP_1;
      break;
    }
  }

  return 0;
}

static int
build_grid_haven(world *wp, int z)
{
  if (wp == NULL)
    return 1;
  if ((z < 0) || (z >= wp->size_z))
    return 1;
  if ((wp->size_x < 7) || wp->size_y < 7)
    return 1;

  fill_floor(wp, z, 0, 0, wp->size_x - 1, wp->size_y - 1, GR_WALL);

  fill_floor(wp, z,
             wp->size_x / 2 - 2, wp->size_y / 2 - 2,
             wp->size_x / 2 + 2, wp->size_y / 2 + 2,
             GR_FLOOR);

  if (z == wp->size_z - 1)
    wp->grid[z][wp->size_x / 2 - 2][wp->size_y / 2] = GR_GATE_OF_DOOM;
  else if (z <= wp->size_z - 2)
    wp->grid[z][wp->size_x / 2 - 2][wp->size_y / 2] = GR_STAIR_DOWN_0;

  if (z <= wp->size_z - 2)
    wp->grid[z][wp->size_x / 2 + 2][wp->size_y / 2] = GR_STAIR_DOWN_1;

  if (z >= 1)
    wp->grid[z][wp->size_x / 2][wp->size_y / 2 - 2] = GR_STAIR_UP_0;

  if (z >= 1)
    wp->grid[z][wp->size_x / 2][wp->size_y / 2 + 2] = GR_STAIR_UP_1;

  return 0;
}

static int
build_grid_rubble(world *wp, int z)
{
  int i;
  int m;
  int n;
  int x;
  int y;
  int type;

  if (wp == NULL)
    return 1;
  if ((z < 0) || (z >= wp->size_z))
    return 1;

  if ((wp->size_x < 9) || wp->size_y < 9)
    return 1;

  fill_floor(wp, z, 0, 0, wp->size_x - 1, wp->size_y - 1, GR_WALL);

  fill_floor(wp, z, 1, 1, wp->size_x - 2, wp->size_y - 2, GR_FLOOR);

  for (i = 0; i < wp->size_x * wp->size_y / 10; i++)
  {
    do
    {
      x = 1 + (rand() % (wp->size_x - 2));
      y = 1 + (rand() % (wp->size_y - 2));
    } while(wp->grid[z][x][y] != GR_FLOOR);
    wp->grid[z][x][y] = GR_WALL;
  }

  n = wp->size_x;
  if (n > wp->size_y)
    n = wp->size_y;
  for (i = 1; i <= n - 2; i++)
  {
    wp->grid[z][i][i] = GR_FLOOR;
    wp->grid[z][i][n - 1 - i] = GR_FLOOR;
  }

  for (i = -3; i <= 3; i++)
    wp->grid[z][n / 2 + i][n / 2] = GR_FLOOR;

  if (z >= 1)
  {
    wp->grid[z][n / 2 - 3][n / 2] = GR_STAIR_UP_0;
    wp->grid[z][n / 2 + 3][n / 2] = GR_STAIR_UP_1;
  }

  m = rand() % 4;
  for (i = 0; i < 4; i++)
  {
    switch (i)
    {
    case 0:
      x = 2;
      y = 2;
      break;
    case 1:
      x = n - 3;
      y = 2;
      break;
    case 2:
      x = n - 3;
      y = n - 3;
      break;
    default:
      x = 2;
      y = n - 3;
      break;
    }

    if ((i + m) % 4 == 0)
    {
      if (z == wp->size_z - 1)
        type = GR_GATE_OF_DOOM;
      else if (z <= wp->size_z - 2)
        type = GR_STAIR_DOWN_0;
      else
        continue;
    }
    else if ((i + m) % 4 == 2)
    {
      if (z <= wp->size_z - 2)
        type = GR_STAIR_DOWN_1;
      else
        continue;
    }
    else
    {
      continue;
    }

    wp->grid[z][x][y] = type;
  }

  return 0;
}

static int
build_grid_glass_castle(world *wp, int z)
{
  int i;
  int x;
  int y;
  int found;

  if (wp == NULL)
    return 1;
  if ((z < 0) || (z >= wp->size_z))
    return 1;

  if ((wp->size_x < 20) || wp->size_y < 20)
    return 1;

  fill_floor(wp, z, 0, 0, wp->size_x - 1, wp->size_y - 1, GR_WALL);

  build_grid_bsp_recursive(wp, z,
                           1, 1,
                           wp->size_x - 2, wp->size_y - 2,
                           0, 5);

  found = 0;
  for (x = wp->size_x / 4 - 1; x <= (wp->size_x * 3) / 4 + 1; x++)
  {
    for (y = wp->size_y / 4 - 1; y <= (wp->size_y * 3) / 4 + 1; y++)
    {
      if (!grid_blocks_creature(wp, z, x, y))
      {
        found = 1;
        break;
      }
    }
    if (found)
      break;
  }

  if (found)
  {
    fill_floor(wp, z,
               wp->size_x / 4 - 1, wp->size_y / 4 - 1,
               (wp->size_x * 3) / 4 + 1, (wp->size_y * 3) / 4 + 1,
               GR_FLOOR);
    fill_floor(wp, z,
               wp->size_x / 4, wp->size_y / 4,
               (wp->size_x * 3) / 4, (wp->size_y * 3) / 4,
               GR_GLASS_WALL);
    build_grid_bsp_recursive(wp, z,
                             wp->size_x / 4 + 1, wp->size_y / 4 + 1,
                             (wp->size_x * 3) / 4 - 1,
                             (wp->size_y * 3) / 4 - 1,
                             0, 4);
    wp->grid[z][wp->size_x / 4 + 1][wp->size_y / 4] = GR_FLOOR;
    wp->grid[z][wp->size_x / 4][wp->size_y / 4 + 1] = GR_FLOOR;
    wp->grid[z][(wp->size_x * 3) / 4 - 1][wp->size_y / 4] = GR_FLOOR;
    wp->grid[z][(wp->size_x * 3) / 4][wp->size_y / 4 + 1] = GR_FLOOR;
    wp->grid[z][wp->size_x / 4 + 1][(wp->size_y * 3) / 4] = GR_FLOOR;
    wp->grid[z][wp->size_x / 4][(wp->size_y * 3) / 4 - 1] = GR_FLOOR;
    wp->grid[z][(wp->size_x * 3) / 4 - 1][(wp->size_y * 3) / 4] = GR_FLOOR;
    wp->grid[z][(wp->size_x * 3) / 4][(wp->size_y * 3) / 4 - 1] = GR_FLOOR;
  }

  for (i = 0; i < 4; i++)
  {
    while (1 == 1)
    {
      x = 1 + (rand() % (wp->size_x - 2));
      y = 1 + (rand() % (wp->size_y - 2));
      if (wp->grid[z][x][y] != GR_FLOOR)
        continue;
      if (!found)
        break;
      if ((x >= wp->size_x / 4 - 1) && (x <= (wp->size_x * 3) / 4 + 1)
          && (y >= wp->size_y / 4 - 1) && (y <= (wp->size_y * 3) / 4 + 1))
      {
        if ((x <= wp->size_x / 4)
            || (x >= (wp->size_x * 3) / 4)
            || (y <= wp->size_y / 4)
            || (y >= (wp->size_y * 3) / 4))
          continue;
        if (i < 2)
          continue;
      }
      else
      {
        if (i >= 2)
          continue;
      }
      break;
    }

    switch (i)
    {
    case 0:
      if (z >= 1)
        wp->grid[z][x][y] = GR_STAIR_UP_0;
      break;
    case 1:
      if (z >= 1)
        wp->grid[z][x][y] = GR_STAIR_UP_1;
      break;
    case 2:
      if (z == wp->size_z - 1)
        wp->grid[z][x][y] = GR_GATE_OF_DOOM;
      else if (z <= wp->size_z - 2)
        wp->grid[z][x][y] = GR_STAIR_DOWN_0;
      break;
    default:
      if (z <= wp->size_z - 2)
        wp->grid[z][x][y] = GR_STAIR_DOWN_1;
      break;
      break;
    }
  }

  return 0;
}

int
build_new_game(world *wp)
{
  int i;
  int n;
  int z;
  int x;
  int y;
  int num_scroll;

  if (wp == NULL)
  {
    fprintf(stderr, "build_new_game: wp is NULL");
    return 1;
  }

  world_reset_game(wp);
  randomize_int_array(wp->label, NUM_SCROLL);

  for (z = 0; z < wp->size_z; z++)
  {
    if (z == 0)
      build_grid_haven(wp, z);
    else if (z == wp->size_z - 1)
      build_grid_rubble(wp, z);
    else if (z == wp->size_z - 2)
      build_grid_glass_castle(wp, z);
    else
      build_grid_bsp(wp, z);
  }

  n = add_new_creature(wp, CR_PLAYER,
                       0, wp->size_x / 2, wp->size_y / 2, 0);
  if (n < 0)
  {
    fprintf(stderr, "build_new_game: add_new_creature(player) failed\n");
    return 1;
  }
  wp->player_id = n;
  wp->turn_z = wp->cr[n]->z;
  wp->camera_z = wp->cr[n]->z;
  wp->camera_x = wp->cr[n]->x;
  wp->camera_y = wp->cr[n]->y;
  wp->cursor_x = wp->cr[n]->x;
  wp->cursor_y = wp->cr[n]->y;

  /* no monster or scroll in B0F --- it is too small */
  for (z = 1; z < wp->size_z; z++)
  {
    for (i = 0; i < 48; i++)
    {
      do
      {
        x = 1 + (rand() % (wp->size_x - 2));
        y = 1 + (rand() % (wp->size_y - 2));
      } while ((grid_blocks_creature(wp, z, x, y))
               || (grid_has_creature(wp, z, x, y)));
      add_new_creature(wp, random_creature_type(z),
                       z, x, y, 1 + (i % 2));
    }

    num_scroll = 7 + (rand() % 3);
    for (i = 0; i < num_scroll; i++)
    {
      do
      {
        x = 1 + (rand() % (wp->size_x - 2));
        y = 1 + (rand() % (wp->size_y - 2));
      } while (wp->grid[z][x][y] != GR_FLOOR);
      n = 1;
      if (i % 4 == 0)
        n = 2;
      add_new_object(wp, random_scroll_type(z), z, x, y, n);
    }
  }

  return 0;
}
