/* $Id: loop.c,v 1.434 2010/03/27 15:56:41 oohara Exp $ */

#include <stdio.h>
/* free, qsort, rand */
#include <stdlib.h>
#include <curses.h>
/* tigetnum */
#include <term.h>

#include "world.h"
#include "dungeon.h"
#include "creature.h"
#include "action.h"
#include "monster.h"
#include "los.h"
#include "creature-detail.h"
#include "grid-detail.h"
#include "scroll-detail.h"
#include "enchantment-detail.h"

#include "loop.h"

static creature *find_creature_under_cursor(world *wp);
static int cursor_next_target(world *wp);
static int examine_stair_destination(world *wp, int reset_only);
static int loop_input_0(world *wp);
static int loop_input_1(world *wp);
static int loop_input_2(world *wp);
static int loop_input_3(world *wp);
static int loop_input_4(world *wp);
static int loop_input_5(world *wp);
static int loop_input_6(world *wp);
static int loop_input_7(world *wp);
static int loop_input_8(world *wp);
static int loop_input(world *wp);
static int remove_dead(world *wp);
static int check_victory(world *wp);
static int check_shadow_los(world *wp);
static int check_candle(world *wp);
static int player_remember_map(world *wp);
static int creature_respawn(world *wp);
static int loop_upkeep_action(world *wp);
static int loop_upkeep_turn(world *wp, int creature_id);
static int loop_upkeep_ut(world *wp, int ut);
static int loop_action(world *wp);
static int show_multiline(world *wp, int y_top, const char *str);
static int show_log(world *wp, int y_top, int y_bottom, int log_bottom);
static int loop_draw_0_how_to_show_this_grid(world *wp, int x, int y);
static int loop_draw_0_grid(world *wp);
static int loop_draw_0_object(world *wp);
static int loop_draw_0_creature(world *wp);
static int loop_draw_0_stat_one(world *wp, int creature_id, int y);
static int loop_draw_0_stat(world *wp);
static int loop_draw_0_message_0(world *wp);
static int loop_draw_0_message_1(world *wp);
static int loop_draw_0_message_2(world *wp);
static int loop_draw_0_message_7(world *wp);
static int loop_draw_0_message_8(world *wp);
static int loop_draw_0_message(world *wp);
static int loop_draw_0(world *wp);
static int loop_draw_1(world *wp);
static int loop_draw_3(world *wp);
static int loop_draw_4(world *wp);
static int loop_draw_5(world *wp);
static int loop_draw_6(world *wp);
static int loop_draw(world *wp);

/* list of wp->mode[0]
 * 0: waiting for command
 * 1: inventory
 *    [1] inventory chosen
 *    [2] page
 *    [3] 0: examining
 *        1: using
 * 2: cursor
 * 3: message log
 *    [1] log bottom
 * 4: desribe scroll
 *    [1] type
 * 5: dscribe creature
 *    [1] creature id
 * 6: full-screen map
 *    [1] up-left corner x
 *    [2] up-left corner y
 * 7: cast shadow
 * 8: resign
 */

static
creature *find_creature_under_cursor(world *wp)
{
  int i;
  creature *who = NULL;

  if (wp == NULL)
    return NULL;

  who = NULL;
  for (i = wp->cr_size - 1; i >= 0; i--)
  {
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->z != wp->camera_z)
      continue;
    if ((wp->cr[i]->x == wp->cursor_x)
        && (wp->cr[i]->y == wp->cursor_y))
      who = wp->cr[i];
  }

  return who;
}

static int
cursor_next_target(world *wp)
{
  int i;
  int x;
  int y;
  int n;
  int n0;
  creature *who = NULL;

  if (wp == NULL)
    return 1;

  who = find_creature_under_cursor(wp);
  if (who == NULL)
    n0 = wp->player_id;
  else
    n0 = who->id;

  for (i = 1; i < wp->cr_size; i++)
  {
    n = (n0 + i) % wp->cr_size;
    if (n == wp->player_id)
      continue;
    if (creature_is_dead(wp, n))
      continue;
    if (wp->cr[n]->z != wp->camera_z)
      continue;
    x = wp->cr[n]->x;
    y = wp->cr[n]->y;
    if (distance(x, y, wp->camera_x, wp->camera_y) > 8)
      continue;
    wp->cursor_x = x;
    wp->cursor_y = y;
    break;
  }

  return 0;
}

static int
examine_stair_destination(world *wp, int reset_only)
{
  int dest_z;
  int dest_x;
  int dest_y;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
    return 1;
  if (wp->cr[wp->player_id] == NULL)
    return 1;
  player = wp->cr[wp->player_id];

  if (wp->camera_z != player->z)
  {
    wp->camera_z = player->z;
    wp->camera_x = player->x;
    wp->camera_y = player->y;
    wp->cursor_x = player->x;
    wp->cursor_y = player->y;
    return 0;
  }

  if (reset_only)
    return 0;

  if (!has_los_from_player(wp, wp->camera_z, wp->cursor_x, wp->cursor_y))
    return 0;
  if (!find_stair_destination(wp,
                              wp->camera_z, wp->cursor_x, wp->cursor_y,
                              &dest_z, &dest_x, &dest_y))
    return 0;

  wp->camera_z = dest_z;
  wp->camera_x = dest_x;
  wp->camera_y = dest_y;
  wp->cursor_x = dest_x;
  wp->cursor_y = dest_y;

  return 0;
}

static int
loop_input_0(world *wp)
{
  int c;
  int i;
  int scr_size_x;
  int scr_size_y;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  scr_size_x = tigetnum("cols");
  scr_size_y = tigetnum("lines");

  c = getch();

  switch (c)
  {
  case 'Q':
    if ((creature_is_dead(wp, wp->player_id))
        || (wp->victory))
    {
      wp->should_quit = 1;
      return 0;
    }
    else
    {
      wp->mode[0] = 8;
    }
    break;
  case ' ':
    cursor_next_target(wp);
    break;
  case 'i':
    sort_inventory(wp, 2);
    wp->mode[0] = 1;
    wp->mode[1] = -1;
    wp->mode[2] = 0;
    wp->mode[3] = 0;
    break;
  case 'x':
    wp->mode[0] = 2;
    break;
  case 'p':
    wp->mode[0] = 3;
    wp->mode[1] = 0;
    break;
  case 'm':
    wp->mode[0] = 6;
    wp->mode[1] = wp->camera_x - scr_size_x / 2;
    wp->mode[2] = wp->camera_y - scr_size_y / 2;
    break;
  default:
    break;
  }

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];

  switch (c)
  {
  case '7':
  case 'y':
    player->act[0] = ACTION_WALK;
    player->act[1] = -1;
    player->act[2] = -1;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    player->act[0] = ACTION_WALK;
    player->act[1] = 0;
    player->act[2] = -1;
    break;
  case '9':
  case 'u':
    player->act[0] = ACTION_WALK;
    player->act[1] = 1;
    player->act[2] = -1;
    break;
  case '4':
  case 'h':
  case KEY_LEFT:
    player->act[0] = ACTION_WALK;
    player->act[1] = -1;
    player->act[2] = 0;
    break;
  case '5':
  case '.':
    player->act[0] = ACTION_STAY;
    break;
  case '6':
  case 'l':
  case KEY_RIGHT:
    player->act[0] = ACTION_WALK;
    player->act[1] = 1;
    player->act[2] = 0;
    break;
  case '1':
  case 'b':
    player->act[0] = ACTION_WALK;
    player->act[1] = -1;
    player->act[2] = 1;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    player->act[0] = ACTION_WALK;
    player->act[1] = 0;
    player->act[2] = 1;
    break;
  case '3':
  case 'n':
    player->act[0] = ACTION_WALK;
    player->act[1] = 1;
    player->act[2] = 1;
    break;
  case 'S':
    player->act[0] = ACTION_SAVE;
    break;
  case '<':
  case '>':
    player->act[0] = ACTION_CLIMB_STAIR;
    break;
  case 'r':
    sort_inventory(wp, 0);
    wp->mode[0] = 1;
    wp->mode[1] = -1;
    wp->mode[2] = 0;
    wp->mode[3] = 1;
    break;
  case 't':
    sort_inventory(wp, 1);
    wp->mode[0] = 1;
    wp->mode[1] = -1;
    wp->mode[2] = 0;
    wp->mode[3] = 1;
    break;
  case 's':
    wp->mode[0] = 7;
    break;
  case 'Z':
    player->act[0] = ACTION_DESTROY_SHADOW;
    break;
  case 'a':
    wp->shadow_mode = (wp->shadow_mode + 1) % 3;
    break;
  default:
    break;
  }

  if (!action_check(wp, player->id, player->act))
  {
    for (i = 0; i < 4; i++)
      player->act[i] = 0;
  }

  return 0;
}

static int
loop_input_1(world *wp)
{
  int c;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'c':
    wp->mode[0] = 0;
    break;
  case 'r':
  case 't':
    if (count_inventory(wp) <= 0)
    {
      wp->mode[0] = 0;
    }
    else if (wp->mode[3] == 1)
    {
      wp->mode[0] = 0;
    }
    else if (!creature_is_dead(wp, wp->player_id))
    {
      wp->mode[3] = 1;
    }
    break;
  case 'i':
    if (count_inventory(wp) <= 0)
    {
      wp->mode[0] = 0;
    }
    else if (wp->mode[3] == 0)
    {
      wp->mode[0] = 0;
    }
    else
    {
      wp->mode[3] = 0;
    }
    break;
  default:
    break;
  }

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];

  if (wp->mode[1] < 0)
  {
    switch (c)
    {
    case '7':
    case 'y':
      wp->mode[1] = wp->mode[2] * 9 + 6;
      break;
    case '8':
    case 'k':
    case KEY_UP:
      wp->mode[1] = wp->mode[2] * 9 + 7;
      break;
    case '9':
    case 'u':
      wp->mode[1] = wp->mode[2] * 9 + 8;
      break;
    case '4':
    case 'h':
      wp->mode[1] = wp->mode[2] * 9 + 3;
      break;
    case '5':
    case '.':
      wp->mode[1] = wp->mode[2] * 9 + 4;
      break;
    case '6':
    case 'l':
    case KEY_RIGHT:
      wp->mode[1] = wp->mode[2] * 9 + 5;
      break;
    case '1':
    case 'b':
      wp->mode[1] = wp->mode[2] * 9 + 0;
      break;
    case '2':
    case 'j':
    case KEY_DOWN:
      wp->mode[1] = wp->mode[2] * 9 + 1;
      break;
    case '3':
    case 'n':
      wp->mode[1] = wp->mode[2] * 9 + 2;
      break;
    case '+':
    case ' ':
      (wp->mode[2])++;
      break;
    case '-':
      (wp->mode[2])--;
      break;
    case '?':
      sort_inventory(wp, 0);
      break;
    case '!':
      sort_inventory(wp, 1);
      break;
    default:
      break;
    }
    if (wp->mode[1] >= 0)
    {      
      if (inventory_is_dead(wp, wp->mode[1]))
      {  
        wp->mode[1] = -1;
      }
      else
      {
        switch (wp->mode[3])
        {
        case 0:
          wp->mode[0] = 4;
          wp->mode[1] = wp->inv[wp->mode[1]]->type;
          break;
        case 1:
          break;
        default:
          wp->mode[1] = -1;
          break;
        }
      }
    }
  }
  else
  {
    switch (c)
    {
    case '7':
    case 'y':
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = -1;
      player->act[3] = -1;
      wp->mode[0] = 0;
      break;
    case '8':
    case 'k':
    case KEY_UP:
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = 0;
      player->act[3] = -1;
      wp->mode[0] = 0;
      break;
    case '9':
    case 'u':
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = 1;
      player->act[3] = -1;
      wp->mode[0] = 0;
      break;
    case '4':
    case 'h':
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = -1;
      player->act[3] = 0;
      wp->mode[0] = 0;
      break;
    case '5':
    case '.':
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = 0;
      player->act[3] = 0;
      wp->mode[0] = 0;
      break;
    case '6':
    case 'l':
    case KEY_RIGHT:
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = 1;
      player->act[3] = 0;
      wp->mode[0] = 0;
      break;
    case '1':
    case 'b':
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = -1;
      player->act[3] = 1;
      wp->mode[0] = 0;
      break;
    case '2':
    case 'j':
    case KEY_DOWN:
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = 0;
      player->act[3] = 1;
      wp->mode[0] = 0;
      break;
    case '3':
    case 'n':
      player->act[0] = ACTION_USE_SCROLL;
      player->act[1] = wp->mode[1];
      player->act[2] = 1;
      player->act[3] = 1;
      wp->mode[0] = 0;
      break;
    default:
      break;
    }
  }

  return 0;
}

static int
loop_input_2(world *wp)
{
  int c;
  int dx;
  int dy;
  creature *who = NULL;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'x':
  case 'c':
    examine_stair_destination(wp, 1);
    wp->mode[0] = 0;
    break;
  case '.':
  case '5':
    who = find_creature_under_cursor(wp);
    if ((who != NULL)
        && (creature_is_face_to_face(wp, who->id)))
    {
      wp->mode[0] = 5;
      wp->mode[1] = who->id;
    }
    break;
  default:
    break;
  }

  if (wp->mode[0] != 2)
    return 0;

  dx = 0;
  dy = 0;
  switch (c)
  {
  case '<':
  case '>':
    examine_stair_destination(wp, 0);
    break;
  case '7':
  case 'y':
    dx = -1;
    dy = -1;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    dx = 0;
    dy = -1;
    break;
  case '9':
  case 'u':
    dx = 1;
    dy = -1;
    break;
  case '4':
  case 'h':
    dx = -1;
    dy = 0;
    break;
  case '5':
  case '.':
    break;
  case '6':
  case 'l':
  case KEY_RIGHT:
    dx = 1;
    dy = 0;
    break;
  case '1':
  case 'b':
    dx = -1;
    dy = 1;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    dx = 0;
    dy = 1;
    break;
  case '3':
  case 'n':
    dx = 1;
    dy = 1;
    break;
  case ' ':
    cursor_next_target(wp);
    break;
  default:
    break;
  }

  if (distance(wp->cursor_x + dx, wp->cursor_y + dy,
               wp->camera_x, wp->camera_y) <= 8)
  {
    wp->cursor_x += dx;
    wp->cursor_y += dy;
  }

  return 0;
}

static int
loop_input_3(world *wp)
{
  int c;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'p':
  case 'c':
    wp->mode[0] = 0;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    wp->mode[1] += 1;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    wp->mode[1] -= 1;
    break;
  case '+':
    wp->mode[1] -= 10;
    break;
  case '-':
  case ' ':
    wp->mode[1] += 10;
    break;
  default:
    break;
  }

  return 0;
}

static int
loop_input_4(world *wp)
{
  int c;
  int n;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'i':
  case 'c':
    wp->mode[0] = 0;
    break;
  default:
    break;
  }

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  if (wp->mode[0] != 4)
    return 0;

  switch (c)
  {
  case 'r':
  case 't':
    n = inventory_find_this_type(wp, wp->mode[1]);
    if (n >= 0)
    {
      wp->mode[0] = 1;
      wp->mode[1] = n;
      wp->mode[2] = 0;
      wp->mode[3] = 1;
    }
    break;
  default:
    break;
  }

  return 0;
}

static int
loop_input_5(world *wp)
{
  int c;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'x':
  case 'c':
    wp->mode[0] = 0;
    break;
  case '.':
  case '5':
    wp->mode[0] = 2;
    break;
  default:
    break;
  }

  return 0;
}

static int
loop_input_6(world *wp)
{
  int c;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'm':
  case 'c':
    wp->mode[0] = 0;
    break;
  case '7':
  case 'y':
    wp->mode[1] += -1;
    wp->mode[2] += -1;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    wp->mode[1] += 0;
    wp->mode[2] += -1;
    break;
  case '9':
  case 'u':
    wp->mode[1] += 1;
    wp->mode[2] += -1;
    break;
  case '4':
  case 'h':
    wp->mode[1] += -1;
    wp->mode[2] += 0;
    break;
  case '5':
  case '.':
    break;
  case '6':
  case 'l':
  case KEY_RIGHT:
    wp->mode[1] += 1;
    wp->mode[2] += 0;
    break;
  case '1':
  case 'b':
    wp->mode[1] += -1;
    wp->mode[2] += 1;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    wp->mode[1] += 0;
    wp->mode[2] += 1;
    break;
  case '3':
  case 'n':
    wp->mode[1] += 1;
    wp->mode[2] += 1;
    break;
  default:
    break;
  }

  return 0;
}

static int
loop_input_7(world *wp)
{
  int i;
  int c;
  creature *player = NULL;
  
  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 's':
  case 'c':
    wp->mode[0] = 0;
    break;
  default:
    break;
  }

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];
  if (wp->mode[0] != 7)
    return 0;

  switch (c)
  {
  case '7':
  case 'y':
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = -1;
    player->act[2] = -1;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = 0;
    player->act[2] = -1;
    break;
  case '9':
  case 'u':
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = 1;
    player->act[2] = -1;
    break;
  case '4':
  case 'h':
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = -1;
    player->act[2] = 0;
    break;
  case '5':
  case '.':
    break;
  case '6':
  case 'l':
  case KEY_RIGHT:
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = 1;
    player->act[2] = 0;
    break;
  case '1':
  case 'b':
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = -1;
    player->act[2] = 1;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = 0;
    player->act[2] = 1;
    break;
  case '3':
  case 'n':
    player->act[0] = ACTION_CREATE_SHADOW;
    player->act[1] = 1;
    player->act[2] = 1;
    break;
  default:
    break;
  }

  if (action_check(wp, player->id, player->act))
  {
    wp->mode[0] = 0;
  }
  else
  {
    for (i = 0; i < 4; i++)
      player->act[i] = 0;
  }

  return 0;
}

static int
loop_input_8(world *wp)
{
  int c;

  if (wp == NULL)
    return 1;

  c = getch();

  switch (c)
  {
  case 'o':
    wp->should_quit = 1;
    return 0;
    break;
  case 'Q':
  case 'c':
    wp->mode[0] = 0;
    break;
  default:
    break;
  }

  return 0;
}

static int
loop_input(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "loop_input: wp is NULL\n");
    return 1;
  }

  switch (wp->mode[0])
  {
  case 0:
    return loop_input_0(wp);
    break;
  case 1:
    return loop_input_1(wp);
    break;
  case 2:
    return loop_input_2(wp);
    break;
  case 3:
    return loop_input_3(wp);
    break;
  case 4:
    return loop_input_4(wp);
    break;
  case 5:
    return loop_input_5(wp);
    break;
  case 6:
    return loop_input_6(wp);
    break;
  case 7:
    return loop_input_7(wp);
    break;
  case 8:
    return loop_input_8(wp);
    break;
  default:
    return 1;
    break;
  }

  return 1;
}

static int
remove_dead(world *wp)
{
  int i;
  char buf[128];
  struct creature_detail crd;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (wp->cr[i] == NULL)
      continue;
    if (creature_is_dead(wp, i))
    {
      if (i != wp->player_id)
        remove_creature(wp, i);
      continue;
    }
    if (wp->cr[i]->enchant_type > 0)
    {
      if (wp->cr[i]->enchant_duration <= 0)
      {
        creature_detail_get(wp->cr[i]->type, &crd);
        sprintf(buf, "%s no longer has (0) %s",
                crd.name, enchantment_name(wp->cr[i]->enchant_type));
        add_log(wp, buf);
        wp->cr[i]->enchant_type = 0;
        wp->cr[i]->enchant_duration = 0;
      }
    }
  }

  for (i = 0; i < wp->obj_size; i++)
  {
    if (wp->obj[i] == NULL)
      continue;
    if (object_is_dead(wp, i))
      remove_object(wp, i);
  }

  for (i = 0; i < wp->inv_size; i++)
  {
    if (inventory_is_dead(wp, i))
    {
      wp->inv[i]->type = -1;
      wp->inv[i]->quantity = 0;
    } 
  }

  return 0;
}

static int
check_victory(world *wp)
{
  int grid_type;
  char buf[128];
  struct creature_detail crd;
  struct grid_detail grd;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];
  creature_detail_get(player->type, &crd);

  if (wp->victory)
    return 0;

  if (grid_is_illegal(wp, player->z, player->x, player->y))
    return 0;
  grid_type = wp->grid[player->z][player->x][player->y] & GRID_TERRAIN_MASK;
  if (grid_type != GR_GATE_OF_DOOM)
    return 0;
  if (wp->candle <= 0)
    return 0;

  grid_detail_get(grid_type, &grd);
  sprintf(buf, "%s closes %s", crd.name, grd.name);
  add_log(wp, buf);
  wp->grid[player->z][player->x][player->y] = GR_WALL;
  player->hp = 0;
  wp->victory = 1;

  return 0;
}

static int
check_shadow_los(world *wp)
{
  int i;
  char buf[128];
  struct creature_detail crd;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (i == wp->player_id)
      continue;
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->type != CR_PLAYER_SHADOW)
      continue;

    if ((creature_is_dead(wp, wp->player_id))
        || (wp->cr[i]->z != wp->cr[wp->player_id]->z)
        || (distance(wp->cr[i]->x, wp->cr[i]->y,
                     wp->cr[wp->player_id]->x, wp->cr[wp->player_id]->y) > 6)
        || (!has_los_from_player(wp,
                                 wp->cr[i]->z,
                                 wp->cr[i]->x,
                                 wp->cr[i]->y)))
    {
      creature_detail_get(wp->cr[i]->type, &crd);
      sprintf(buf, "%s dissolves in the dark", crd.name);
      add_log(wp, buf);
      wp->cr[i]->hp = 0;
    }
  }

  return 0;
}

static int
check_candle(world *wp)
{
  if (wp == NULL)
    return 1;

  if (creature_is_dead(wp, wp->player_id))
    return 0;

  if (wp->candle <= 0)
  {
    add_log(wp, "the candle runs out");
    wp->cr[wp->player_id]->hp = 0;
  }

  return 0;
}

static int
player_remember_map(world *wp)
{
  int z;
  int x;
  int y;
  int dx;
  int dy;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
    return 1;
  if (wp->cr[wp->player_id] == NULL)
    return 1;
  player = wp->cr[wp->player_id];

  z = player->z;
  x = player->x;
  y = player->y;
  if (grid_is_illegal(wp, z, x, y))
    return 1;

  for (dx = -6; dx <= 6; dx++)
  {
    for (dy = -6; dy <= 6; dy++)
    {
      if (grid_is_illegal(wp, z, x + dx, y + dy))
        continue;
      /* has_los() is expensive and should not be called unless necessary */
      if ((wp->grid[z][x + dx][y + dy] & GRID_PLAYER_SEEN)
          == GRID_PLAYER_SEEN)
        continue;
      if (!has_los(wp,
                   z, x, y,
                   z, x + dx, y + dy))
        continue;
      wp->grid[z][x + dx][y + dy] |= GRID_PLAYER_SEEN;
    }
  }

  return 0;
}

static int
creature_respawn(world *wp)
{
  int i;
  int z;
  int x;
  int y;
  int found;
  int wait;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if (creature_is_dead(wp, wp->player_id))
    return 0;
  player = wp->cr[wp->player_id];

  if (wp->turn_z < 1)
    return 0;
  if (list_int_length_all(wp->turn_order[wp->turn_z]) >= 49)
    return 0;

  z = wp->turn_z;
  found = 0;
  for (i = 0; i < 100; i++)
  {
    x = 1 + (rand() % (wp->size_x - 2));
    y = 1 + (rand() % (wp->size_y - 2));
    if (grid_blocks_creature(wp, z, x, y))
      continue;
    if (distance(x, y, player->x, player->y) <= 12)
      continue;
    found = 1;
    break;
  }

  if (!found)
    return 0;

  wait = 1 + (rand() % 2);
  if ((wp->turn_order[wp->turn_z] != NULL)
      && (wp->cr[wp->turn_order[wp->turn_z]->n] != NULL))
    wait += wp->cr[wp->turn_order[wp->turn_z]->n]->wait;
  if (wait < 1)
    wait = 1;

  add_new_creature(wp, random_creature_type(z),
                   z, x, y, wait);

  return 0;
}

static int
loop_upkeep_action(world *wp)
{
  if (wp == NULL)
    return 1;

  player_remember_map(wp);
  check_candle(wp);
  check_victory(wp);
  check_shadow_los(wp);
  remove_dead(wp);

  return 0;
}

static int
loop_upkeep_turn(world *wp, int creature_id)
{
  int in_the_last;
  struct creature_detail crd;
  creature *who = NULL;

  if (wp == NULL)
    return 1;
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
    return 1;
  if (wp->cr[creature_id] == NULL)
    return 1;
  who = wp->cr[creature_id];

  if ((who->enchant_type > 0)
      && (who->enchant_duration > 0))
  {
    creature_detail_get(who->type, &crd);
    switch (who->enchant_type)
    {
    case ENC_POISON:
      creature_take_damage(wp, who->id, (crd.max_hp + 3) / 4,
                           -1, 0);
      break;
    case ENC_REGENERATE:
      creature_heal_hp(wp, who->id, -1, 0);
      break;
    default:
      break;
    }

    (who->enchant_duration)--;
  }

  if (who->id == wp->player_id)
  {
    in_the_last = candle_is_in_the_last(wp);
    (wp->candle)--;
    wp->candle -= creature_count_this_type(wp, CR_PLAYER_SHADOW) * 3;
    if ((!in_the_last)
        && (candle_is_in_the_last(wp)))
    {
      add_log(wp, "the candle is about to run out");
    }
  }

  return 0;
}

static int
loop_upkeep_ut(world *wp, int ut)
{
  int step;
  int step_next;

  if (wp == NULL)
    return 1;

  step = wp->ut / 200;
  step_next = (wp->ut + ut) / 200;
  while (step < step_next)
  {
    creature_respawn(wp);
    step++;
  }

  wp->ut += ut;

  return 0;
}

static int
loop_action(world *wp)
{
  int i;
  int ut;
  int turn_z;
  creature *who = NULL;
  list_int *order = NULL;

  if (wp == NULL)
    return 1;

  do
  {
    loop_upkeep_action(wp);

    if (creature_is_dead(wp, wp->player_id))
      return 0;

    /* remember wp->turn_z here because the action may change it */
    turn_z = wp->turn_z;

    if ((turn_z < 0) || (turn_z >= wp->size_z))
    {
      fprintf(stderr, "loop_action: strange turn_z (%d)\n", turn_z);
      return 1;
    }
    if (wp->turn_order[turn_z] == NULL)
      return 0;
    who = wp->cr[wp->turn_order[turn_z]->n];
    if (who == NULL)
    {
      fprintf(stderr, "loop_action: who is NULL (%d)\n",
              wp->turn_order[turn_z]->n);
      return 1;
    }

    ut = who->wait;
    for (i = 0; i < wp->cr_size; i++)
    {
      if (wp->cr[i] == NULL)
        continue;
      if (wp->cr[i]->z != turn_z)
        continue;
      wp->cr[i]->wait -= ut;
    }
    loop_upkeep_ut(wp, ut);

    if (who->id != wp->player_id)
    {
      if (action_decide_monster(wp, who->id) != 0)
      {
        fprintf(stderr, "loop_action: action_decide_monster failed\n");
        return 1;
      }
    }

    ut = action_exec(wp, who->id);
    if (ut > 0)
    {
      loop_upkeep_turn(wp, who->id);

      /* wp->turn_order[turn_z] may be NULL because the only creature
       * in this floor may leave this floor
       */
      if ((wp->turn_order[turn_z] != NULL)
          && (wp->turn_order[turn_z]->n == who->id))
      {
        order = wp->turn_order[turn_z];
        wp->turn_order[turn_z] = list_int_unlink(order);
        wp->turn_order[turn_z] = list_int_concat(wp->turn_order[turn_z],
                                                 order);
      }

      who->wait += ut;

      wp->turn_order[turn_z] = list_int_sort(wp->turn_order[turn_z],
                                             (void *) wp,
                                             &compare_wait);
    }
  } while (wp->turn_order[wp->turn_z]->n != wp->player_id);

  loop_upkeep_action(wp);

  return 0;
}

/* does not wrap at window edge */
static int
show_multiline(world *wp, int y_top, const char *str)
{
  int i;
  int y;

  if (wp == NULL)
    return 1;

  if (str == NULL)
    return 0;

  y = y_top;
  move(y, 0);
  for (i = 0; str[i] != '\0'; i++)
  {
    if (str[i] == '\n')
    {
      y++;
      move(y, 0);
      continue;
    }
    addch(str[i]);
  }

  return 0;
}

static int
show_log(world *wp, int y_top, int y_bottom, int log_bottom)
{
  int i;
  int n;
  int y;
  int attr;

  if (wp == NULL)
    return 1;

  attr = COLOR_PAIR(7);
  attron(attr);
  for (i = log_bottom; i < wp->log_size; i++)
  {
    y = y_bottom - (i - log_bottom);

    if (y < y_top)
      break;

    n = (wp->log_head - 1 - i + wp->log_size * 2) % wp->log_size;
    if (wp->log[n] == NULL)
      continue;
    mvaddstr(y, 0, wp->log[n]);
  }
  attroff(attr);

  return 0;
}

/* return
 * 0 if the grid should not be shown
 * 1 if the grid should be shown in dark gray
 * 2 if the grid should be shown in color
 */
static int
loop_draw_0_how_to_show_this_grid(world *wp, int x, int y)
{
  if (wp == NULL)
    return 0;
  if (grid_is_illegal(wp, wp->camera_z, x, y))
    return 0;

  if ((wp->candle > 0)
      && (distance(x, y, wp->camera_x, wp->camera_y) <= 6)
      && (has_los_from_player(wp, wp->camera_z, x, y)))
    return 2;
  if ((wp->grid[wp->camera_z][x][y] & GRID_PLAYER_SEEN) == GRID_PLAYER_SEEN)
    return 1;

  return 0;
}

static int
loop_draw_0_grid(world *wp)
{
  int i;
  int j;
  int x;
  int y;
  int n;
  int attr;
  struct grid_detail grd;

  if (wp == NULL)
    return 1;

  for (i = -8; i <= 8; i++)
  {
    for (j = -8; j <= 8; j++)
    {
      x = wp->camera_x + i;
      y = wp->camera_y + j;
      if (grid_is_illegal(wp, wp->camera_z, x, y))
        continue;
      grid_detail_get(wp->grid[wp->camera_z][x][y] & GRID_TERRAIN_MASK,
                      &grd);
      attr = grd.attr;

      n = loop_draw_0_how_to_show_this_grid(wp, x, y);
      if (n == 0)
        continue;
      else if (n == 1)
      {
        attr = COLOR_PAIR(8) | A_BOLD;
        if ((x == wp->cursor_x)
            && (y == wp->cursor_y))
        {
          attr = COLOR_PAIR(7);
        }
      }
      attron(attr);
      mvaddstr(j + 8, i + 8, grd.symbol);
      attroff(attr);
    }
  }

  return 0;
}

static int
loop_draw_0_object(world *wp)
{
  int i;
  int attr;
  struct scroll_detail scd;
  const char *s = NULL;
  creature *player = NULL;

  if (wp == NULL)
    return 1;

  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
    return 1;
  if (wp->cr[wp->player_id] == NULL)
    return 1;
  player = wp->cr[wp->player_id];

  if (wp->candle <= 0)
    return 0;

  for (i = 0; i < wp->obj_size; i++)
  {
    if (object_is_dead(wp, i))
      continue;
    if (wp->obj[i]->z != wp->camera_z)
      continue;
    if (distance(wp->camera_x, wp->camera_y,
                 wp->obj[i]->x, wp->obj[i]->y) > 6)
      continue;
    if (!has_los_from_player(wp, wp->obj[i]->z, wp->obj[i]->x, wp->obj[i]->y))
      continue;
    scroll_detail_get(wp->obj[i]->type, &scd);
    if (scd.healthy)
      s = "?";
    else
      s = "!";
    attr = COLOR_PAIR(7);
    attron(attr);
    mvaddstr(wp->obj[i]->y - wp->camera_y + 8,
             wp->obj[i]->x - wp->camera_x + 8,
             s);
    attroff(attr);
  }

  return 0;
}

static int
loop_draw_0_creature(world *wp)
{
  int i;
  int attr;
  const char *s = NULL;
  struct creature_detail crd;

  if (wp == NULL)
    return 1;

  for (i = 0; i < wp->cr_size; i++)
  {
    if (creature_is_dead(wp, i))
      continue;
    if (wp->cr[i]->z != wp->camera_z)
      continue;
    if (distance(wp->cr[i]->x, wp->cr[i]->y,
                 wp->camera_x, wp->camera_y) > 8)
      continue;

    creature_detail_get(wp->cr[i]->type, &crd);
    attr = COLOR_PAIR(7);
    s = crd.symbol;

    if (wp->cr[i]->id == wp->player_id)
    {
      switch (wp->grid[wp->cr[i]->z][wp->cr[i]->x][wp->cr[i]->y]
              & GRID_TERRAIN_MASK)
      {
      case GR_STAIR_DOWN_0:
      case GR_STAIR_DOWN_1:
        attr = COLOR_PAIR(4);
        break;
      case GR_STAIR_UP_0:
      case GR_STAIR_UP_1:
        attr = COLOR_PAIR(2);
        break;
      default:
        break;
      }
      attr |= A_BOLD;
    }

    if (!creature_is_face_to_face(wp, i))
    {
      attr = COLOR_PAIR(8) | A_BOLD;
      s = crd.warning;
      if ((wp->cr[i]->x == wp->cursor_x)
          && (wp->cr[i]->y == wp->cursor_y))
      {
        attr = COLOR_PAIR(7);
      }
    }

    attron(attr);
    mvaddstr(wp->cr[i]->y - wp->camera_y + 8,
             wp->cr[i]->x - wp->camera_x + 8,
             s);
    attroff(attr);
  }

  return 0;
}

static int
loop_draw_0_stat_one(world *wp, int creature_id, int y)
{
  int attr;
  struct creature_detail crd;
  char buf[128];
  creature *who = NULL;

  if (wp == NULL)
    return 1;
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
    return 1;

  if (wp->cr[creature_id] == NULL)
    return 0;
  if ((creature_id != wp->player_id)
      && (creature_is_dead(wp, creature_id)))
    return 0;

  who = wp->cr[creature_id];
  creature_detail_get(who->type, &crd);

  attr = COLOR_PAIR(7);
  attron(attr);
  if (!creature_is_face_to_face(wp, who->id))
  {
    mvaddstr(y, 20, "something");
  }
  else
  { 
    mvaddstr(y, 20, crd.name);
    sprintf(buf, "HP: %2d/%2d", who->hp, crd.max_hp);
    mvaddstr(y + 1, 20, buf);
    sprintf(buf, "power: %2d,%2d",
            crd.power_melee, crd.power_missile);
    mvaddstr(y + 2, 20, buf);
    if ((who->enchant_type > 0)
        && (who->enchant_duration > 0))
    {
      sprintf(buf, "(%d) %s",
              who->enchant_duration, enchantment_name(who->enchant_type));
      mvaddstr(y + 3, 20, buf);
    }
  }
  attroff(attr);

  return 0;
}

static int
loop_draw_0_stat(world *wp)
{
  int attr;
  char buf[128];
  creature *who = NULL;

  if (wp == NULL)
    return 1;

  attr = COLOR_PAIR(7);
  attron(attr);

  sprintf(buf, "B%2dF", wp->camera_z);
  mvaddstr(0, 20, buf);

  attr = COLOR_PAIR(7);
  if (wp->candle <= 0)
    attr = COLOR_PAIR(8) | A_BOLD;
  else if (candle_is_in_the_last(wp))
    attr = COLOR_PAIR(3) | A_BOLD;
  attron(attr);
  sprintf(buf, "candle: %4d", wp->candle);
  mvaddstr(1, 20, buf);
  attroff(attr);

  switch(wp->shadow_mode)
  {
  case 0:
    mvaddstr(2, 20, "=&>");
    break;
  case 1:
    mvaddstr(2, 20, ">&<");
    break;
  default:
    mvaddstr(2, 20, ".&.");
    break;
  }
  sprintf(buf, "shadow: %d",
          creature_count_this_type(wp, CR_PLAYER_SHADOW));
  mvaddstr(2, 24, buf);
  attroff(attr);

  /* useful for debug */
  if (0)
  {
    if (wp->cr[wp->player_id] != NULL)
    {
      attr = COLOR_PAIR(7);
      attron(attr);
      sprintf(buf, "(%d, %d, %d)",
              wp->cr[wp->player_id]->z,
              wp->cr[wp->player_id]->x,
              wp->cr[wp->player_id]->y);
      mvaddstr(3, 20, buf);
      attroff(attr);
    }
  }

  loop_draw_0_stat_one(wp, wp->player_id, 4);

  who = find_creature_under_cursor(wp);
  if (who != NULL)
    loop_draw_0_stat_one(wp, who->id, 9);

  return 0;
}

static int
loop_draw_0_message_0(world *wp)
{
  int y_bottom;
  int attr;
  int scr_size_y;

  if (wp == NULL)
    return 1;

  scr_size_y = tigetnum("lines");

  y_bottom = scr_size_y - 2;
  if (wp->victory)
    y_bottom -= 2;
  else if (creature_is_dead(wp, wp->player_id))
    y_bottom -= 2;

  show_log(wp, 18, y_bottom, 0);

  if (wp->victory)
  {
    attr = COLOR_PAIR(6);
    attron(attr);
    mvaddstr(scr_size_y - 3, 0,
             "The Gate of Doom is closed.  You win the game!");
    mvaddstr(scr_size_y - 2, 0,
             "([x] examine, [i] inventory, [p] log, [m] map, [Q] quit)");
    attroff(attr);
  }
  else if (creature_is_dead(wp, wp->player_id))
  {
    attr = COLOR_PAIR(6);
    attron(attr);
    mvaddstr(scr_size_y - 3, 0, "You die.");
    mvaddstr(scr_size_y - 2, 0,
             "([x] examine, [i] inventory, [p] log, [m] map, [Q] quit)");
    attroff(attr);
  }

  return 0;
}

static int
loop_draw_0_message_1(world *wp)
{
  int attr;
  int scr_size_y;
  char buf[128];
  char *sname = NULL;

  if (wp == NULL)
    return 1;

  scr_size_y = tigetnum("lines");

  sname = scroll_name(wp, wp->inv[wp->mode[1]]->type);
  sprintf(buf, "Use a scroll of %s in which direction?", sname);
  if (sname != NULL)
  {
    free(sname);
    sname = NULL;
  }
  attr = COLOR_PAIR(6);
  attron(attr);
  mvaddstr(18, 0, buf);
  mvaddstr(19, 0, "([r/t/c] cancel)");
  attroff(attr);

  return 0;
}

static int
loop_draw_0_message_2(world *wp)
{
  int z;
  int x;
  int y;
  int attr;
  int scr_size_y;
  struct grid_detail grd;

  if (wp == NULL)
    return 1;

  scr_size_y = tigetnum("lines");

  z = wp->camera_z;
  x = wp->cursor_x;
  y = wp->cursor_y;

  attr = COLOR_PAIR(6);
  attron(attr);
  mvaddstr(18, 0, "Move the cursor to exaimine.");
  attroff(attr);

  if ((!grid_is_illegal(wp, z, x, y))
      && (loop_draw_0_how_to_show_this_grid(wp, x, y) > 0))
  {
    grid_detail_get(wp->grid[z][x][y] & GRID_TERRAIN_MASK, &grd);
    attr = COLOR_PAIR(7);
    attron(attr);
    mvaddstr(19, 0, grd.name);
    attroff(attr);
  }

  attr = COLOR_PAIR(6);
  attron(attr);
  mvaddstr(20, 0, "([space] target, [./5] describe monster, [x/c] cancel)");
  attroff(attr);

  if ((wp->player_id >= 0) && (wp->player_id < wp->cr_size)
      && (wp->cr[wp->player_id] != NULL)
      && (wp->cr[wp->player_id]->z != wp->camera_z))
  {
    attr = COLOR_PAIR(6);
    attron(attr);
    mvaddstr(21, 0, "([</>] return to current floor)");
    attroff(attr);
  }
  else if ((has_los_from_player(wp, z, x, y))
           && (find_stair_destination(wp,
                                      z, x, y,
                                      NULL, NULL, NULL)))
  {
    attr = COLOR_PAIR(6);
    attron(attr);
    mvaddstr(21, 0, "([</>] examine stair destination)");
    attroff(attr);
  }

  return 0;
}

static int
loop_draw_0_message_7(world *wp)
{
  int attr;

  if (wp == NULL)
    return 1;

  attr = COLOR_PAIR(6);
  attron(attr);
  if (creature_count_this_type(wp, CR_PLAYER_SHADOW) >= 2)
  {
    mvaddstr(18, 0, "You can't create that many shadows.");
  }
  else
  { 
    mvaddstr(18, 0, "Create a shadow in which direction?");
  }
  mvaddstr(19, 0, "([s/c] cancel)");
  attroff(attr);

  return 0;
}

static int
loop_draw_0_message_8(world *wp)
{
  int attr;

  if (wp == NULL)
    return 1;

  attr = COLOR_PAIR(6);
  attron(attr);
  mvaddstr(18, 0, "Do you really want to resign the game?.");
  mvaddstr(19, 0, "([o] confirm SUICIDE, [Q/c] cancel)");
  attroff(attr);

  return 0;
}

static int
loop_draw_0_message(world *wp)
{
  if (wp == NULL)
    return 1;

  switch (wp->mode[0])
  {
  case 0:
    loop_draw_0_message_0(wp);
    break;
  case 1:
    loop_draw_0_message_1(wp);
    break;
  case 2:
    loop_draw_0_message_2(wp);
    break;
  case 7:
    loop_draw_0_message_7(wp);
    break;
  case 8:
    loop_draw_0_message_8(wp);
    break;
  default:
    break;
  }

  return 0;
}

static int
loop_draw_0(world *wp)
{
  if (wp == NULL)
    return 1;

  if (distance(wp->cursor_x, wp->cursor_y,
               wp->camera_x, wp->camera_y) > 8)
  {
    if (!creature_is_dead(wp, wp->player_id))
    {
      wp->cursor_x = wp->cr[wp->player_id]->x;
      wp->cursor_y = wp->cr[wp->player_id]->y;
    }
  }
  if (distance(wp->cursor_x, wp->cursor_y,
               wp->camera_x, wp->camera_y) > 8)
  { 
    wp->cursor_x = wp->camera_x;
    wp->cursor_y = wp->camera_y;
  }

  erase();

  loop_draw_0_grid(wp);
  loop_draw_0_object(wp);
  loop_draw_0_creature(wp);
  loop_draw_0_stat(wp);
  loop_draw_0_message(wp);
  
  move(wp->cursor_y - wp->camera_y + 8,
       wp->cursor_x - wp->camera_x + 8);

  refresh();

  return 0;
}

static int
loop_draw_1(world *wp)
{
  int i;
  int y;
  int n;
  int count;
  int attr;
  char buf[128];
  char *sname = NULL;

  if (wp == NULL)
    return 1;

  count = count_inventory(wp);
  if (count <= 0)
    wp->mode[2] = 0;
  else if (wp->mode[2] < 0)
    wp->mode[2] = (count - 1) / 9;
  else if (wp->mode[2] * 9 >= count)
    wp->mode[2] = 0;

  erase();

  attr = COLOR_PAIR(7);
  attron(attr);
  sprintf(buf, "inventory (page %d/%d)",
          wp->mode[2] + 1,
          (count - 1) / 9 + 1);
  mvaddstr(0, 0, buf);
  attroff(attr);

  for (i = 0; i < 9; i++)
  {
    y = i + 2;
    attr = COLOR_PAIR(7);
    attron(attr);
    switch (i)
    {
    case 0:
      mvaddstr(y, 0, "[b/1]");
      break;
    case 1:
      mvaddstr(y, 0, "[j/2]");
      break;
    case 2:
      mvaddstr(y, 0, "[n/3]");
      break;
    case 3:
      mvaddstr(y, 0, "[h/4]");
      break;
    case 4:
      mvaddstr(y, 0, "[./5]");
      break;
    case 5:
      mvaddstr(y, 0, "[n/6]");
      break;
    case 6:
      mvaddstr(y, 0, "[y/7]");
      break;
    case 7:
      mvaddstr(y, 0, "[k/8]");
      break;
    case 8:
      mvaddstr(y, 0, "[u/9]");
      break;
    default:
      break;
    }
    attroff(attr);

    n = wp->mode[2] * 9 + i;
    if (inventory_is_dead(wp, n))
      continue;

    attr = COLOR_PAIR(7);
    if (scroll_is_unknown(wp, wp->inv[n]->type))
      attr = COLOR_PAIR(8) | A_BOLD;
    attron(attr);

    sprintf(buf, "%2d", wp->inv[n]->quantity);
    mvaddstr(y, 6, buf);

    sname = scroll_name(wp, wp->inv[n]->type);
    mvaddstr(y, 10, sname);
    if (sname != NULL)
    {
      free(sname);
      sname = NULL;
    }

    attroff(attr);
  }

  attr = COLOR_PAIR(6);
  attron(attr);

  if (count > 0)
  {  
    sprintf(buf, "%s which scroll?",
            (wp->mode[3] == 0) ? "Examine" : "Use");
    mvaddstr(12, 0, buf);

    if (count > 9)
      mvaddstr(13, 0, "([+/space] next page, [-] previous page)");

    mvaddstr(14, 0, "([?] healthy first, [!] harmful first)");
    switch (wp->mode[3])
    {
    case 0:
      sprintf(buf, "(%s[i/c] cancel)",
              (creature_is_dead(wp, wp->player_id)) ? "" : "[r/t] use, ");
      mvaddstr(15, 0, buf);
      break;
    case 1:
      mvaddstr(15, 0, "([i] examine, [r/t/c] cancel)");
      break;
    default:
      break;
    }
  }
  else
  {
    mvaddstr(12, 0, "You have no scroll.");
    mvaddstr(13, 0, "([i/r/t/c] cancel)");
  }

  attroff(attr);

  move(16, 0);
  refresh();

  return 0;
}

static int
loop_draw_3(world *wp)
{
  int attr;
  int scr_size_y;
  int line_top;
  char buf[128];

  if (wp == NULL)
    return 1;

  scr_size_y = tigetnum("lines");

  if (wp->mode[1] > wp->log_size - (scr_size_y - 3) - 1)
    wp->mode[1] = wp->log_size - (scr_size_y - 3) - 1;
  if (wp->mode[1] < 0)
    wp->mode[1] = 0;

  erase();

  show_log(wp, 0, scr_size_y - 4, wp->mode[1]);

  attr = COLOR_PAIR(6);
  attron(attr);
  line_top = wp->log_size - wp->mode[1] - (scr_size_y - 3);
  if (line_top < 1)
    line_top = 1;
  sprintf(buf, "message log (line %d -- %d)",
          wp->log_size - wp->mode[1] - (scr_size_y - 3),
          wp->log_size - wp->mode[1]);
  mvaddstr(scr_size_y - 3, 0, buf);
  mvaddstr(scr_size_y - 2, 0,
           "([k/8] -1, [j/2] +1, [-/space] -10, [+] +10, [p/c] cancel)");
  attroff(attr);

  move(scr_size_y - 1, 0);
  refresh();

  return 0;
}

static int
loop_draw_4(world *wp)
{
  int attr;
  int scr_size_y;
  int can_use;
  struct scroll_detail scd;
  char buf[128];
  char *sname = NULL;
  const char *description = NULL;

  if (wp == NULL)
    return 1;

  scr_size_y = tigetnum("lines");

  can_use = 0;
  if ((!creature_is_dead(wp, wp->player_id))
      && (inventory_find_this_type(wp, wp->mode[1]) >= 0))
    can_use = 1;

  erase();

  attr = COLOR_PAIR(7);
  attron(attr);

  scroll_detail_get(wp->mode[1], &scd);
  sname = scroll_name(wp, wp->mode[1]);

  if (sname != NULL)
  {
    sprintf(buf, "scroll of %s ('%s')",
            sname,
            (scd.healthy) ? "?" : "!");
    mvaddstr(0, 0, buf);
  }

  if (!scroll_is_unknown(wp, wp->mode[1]))
  {   
    description = "(no description available)";
    if ((scd.description != NULL) &&
        (scd.description[0] != '\0'))
      description = scd.description;

    show_multiline(wp, 2, description);
  }
 
  attroff(attr);

  attr = COLOR_PAIR(6);
  attron(attr);
  sprintf(buf, "(%s[i/c] cancel)",
          (can_use) ? "[r/t] use this scroll, " : "");
  mvaddstr(scr_size_y - 2, 0, buf);
  attroff(attr);

  move(scr_size_y - 1, 0);
  refresh();

  if (sname != NULL)
  {
    free(sname);
    sname = NULL;
  }

  return 0;
}

static int
loop_draw_5(world *wp)
{
  int attr;
  int scr_size_y;
  struct creature_detail crd;
  char buf[128];
  const char *description = NULL;

  if (wp == NULL)
    return 1;

  scr_size_y = tigetnum("lines");

  creature_detail_get(wp->cr[wp->mode[1]]->type, &crd);

  erase();

  attr = COLOR_PAIR(7);
  attron(attr);

  if (!creature_is_face_to_face(wp, wp->mode[1]))
  {
    sprintf(buf, "something ('%s')", crd.warning);
    mvaddstr(0, 0, buf);
  }
  else
  {
    sprintf(buf, "%s ('%s')", crd.name, crd.symbol);
    mvaddstr(0, 0, buf);

    description = "(nothing special)";
    if ((crd.description != NULL)
        && (crd.description[0] != '\0'))
      description = crd.description;

    show_multiline(wp, 2, description);
  }

  attroff(attr);

  attr = COLOR_PAIR(6);
  attron(attr);
  mvaddstr(scr_size_y - 2, 0, "([./5] continue, [x/c] cancel)");
  attroff(attr);

  move(scr_size_y - 1, 0);
  refresh();

  return 0;
}

static int
loop_draw_6(world *wp)
{
  int i;
  int j;
  int z;
  int x;
  int y;
  int scr_size_x;
  int scr_size_y;
  int attr;
  int type;
  struct grid_detail grd;
  struct creature_detail crd;
  char buf[128];
  creature *player = NULL;
  const char *symbol = NULL;

  if (wp == NULL)
    return 1;

  scr_size_x = tigetnum("cols");
  scr_size_y = tigetnum("lines");

  if (wp->mode[1] + scr_size_x / 2 > wp->size_x)
    wp->mode[1] = -(scr_size_x / 2) + wp->size_x;
  if (wp->mode[1] + scr_size_x / 2 < 0)
    wp->mode[1] = -(scr_size_x / 2);
  if (wp->mode[2] + scr_size_y / 2 > wp->size_y)
    wp->mode[2] = -(scr_size_y / 2) + wp->size_y;
  if (wp->mode[2] + scr_size_y / 2 < 0)
    wp->mode[2] = -(scr_size_y / 2);

  player = NULL;
  crd.symbol = "X";
  if ((wp->player_id >= 0) && (wp->player_id < wp->cr_size))
  {  
    player = wp->cr[wp->player_id];
    creature_detail_get(player->type, &crd);
  }

  erase();

  z = wp->camera_z;
  for (i = 0; i <= scr_size_x - 2; i++)
  {
    for (j = 0; j <= scr_size_y - 3; j++)
    {
      x = wp->mode[1] + i;
      y = wp->mode[2] + j;
      if ((z < 0) || (z >= wp->size_z))
        continue;
      if ((x < 0) || (x >= wp->size_x))
        continue;
      if ((y < 0) || (y >= wp->size_y))
        continue;

      if ((wp->grid[z][x][y] & GRID_PLAYER_SEEN) != GRID_PLAYER_SEEN)
        continue;

      type = wp->grid[z][x][y] & GRID_TERRAIN_MASK;
      grid_detail_get(type, &grd);

      attr = COLOR_PAIR(7);
      switch (type)
      {
      case GR_GATE_OF_DOOM:
        attr = COLOR_PAIR(5) | A_BOLD;
        break;
      case GR_STAIR_DOWN_0:
      case GR_STAIR_DOWN_1:
        attr = COLOR_PAIR(4) | A_BOLD;
        break;
      case GR_STAIR_UP_0:
      case GR_STAIR_UP_1:
        attr = COLOR_PAIR(2) | A_BOLD;
        break;
      default:
        break;
      }

      symbol = grd.symbol;
      if ((player != NULL)
          && (z == player->z)
          && (x == player->x)
          && (y == player->y))
        symbol = crd.symbol;

      attron(attr);
      mvaddstr(j, i, symbol);
      attroff(attr);
    }
  }

  attr = COLOR_PAIR(6);
  attron(attr);
  sprintf(buf, "B%2dF ([m/c] cancel)", z);
  mvaddstr(scr_size_y - 2, 0, buf);
  attroff(attr);

  move(scr_size_y - 1, 0);
  refresh();

  return 0;
}

static int
loop_draw(world *wp)
{
  if (wp == NULL)
    return 1;

  switch (wp->mode[0])
  {
  case 0:
    return loop_draw_0(wp);
    break;
  case 1:
    if (wp->mode[1] < 0)
      return loop_draw_1(wp);
    else
      return loop_draw_0(wp);
    break;
  case 2:
    return loop_draw_0(wp);
    break;
  case 3:
    return loop_draw_3(wp);
    break;
  case 4:
    return loop_draw_4(wp);
    break;
  case 5:
    return loop_draw_5(wp);
    break;
  case 6:
    return loop_draw_6(wp);
    break;
  case 7:
    return loop_draw_0(wp);
    break;
  case 8:
    return loop_draw_0(wp);
    break;
  default:
    return 1;
    break;
  }

  return 1;
}

int
main_loop(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "main_loop: wp is NULL\n");
  }

  loop_draw(wp);

  while(1 == 1)
  {
    if (loop_input(wp) != 0)
      break;

    if (loop_action(wp) != 0)
      break;

    if (loop_draw(wp) != 0)
      break;

    if (wp->should_quit)
      break;
  }

  return 0;
}
