/* $Id: main.c,v 1.84 2010/03/27 16:33:27 oohara Exp $ */

#include <stdio.h>
/* srand */
#include <stdlib.h>
/* time */
#include <time.h>
/* strcmp */
#include <string.h>
#include <curses.h>
/* setupterm */
#include <term.h>

#include "world.h"
#include "loop.h"
#include "save.h"
#include "dungeon.h"
#include "build.h"
#include "util.h"
#include "sanity.h"

static int check_window_size(void);
static int do_version(void);
static int do_help(void);

static int
check_window_size(void)
{
  int errret;
  int scr_size_x;
  int scr_size_y;
  int scr_size_x_needed;
  int scr_size_y_needed;

  if (setupterm(NULL, 1, &errret) != OK)
  {
    fprintf(stderr, "check_window_size: setupterm failed\n");
    return 1;
  }

  scr_size_x = tigetnum("cols");
  if (scr_size_x < 0)
  {
    fprintf(stderr, "check_window_size: tigetnum(cols) failed\n");
    return 1;
  }
  scr_size_y = tigetnum("lines");
  if (scr_size_y < 0)
  {
    fprintf(stderr, "check_window_size: tigetnum(lines) failed\n");
    return 1;
  }

  scr_size_x_needed = 80;
  scr_size_y_needed = 24;
  if ((scr_size_x < scr_size_x_needed)
      || (scr_size_y < scr_size_y_needed))
  {
    fprintf(stderr, "The window size %dx%d is too small.\n",
            scr_size_x, scr_size_y);
    fprintf(stderr, "It must be %dx%d or larger.\n",
            scr_size_x_needed, scr_size_y_needed);
    return 1;
  }

  return 0;
}

static int
do_version(void)
{
  printf("last-of-candle (Sun, 28 Mar 2010 release)\n");
  printf("Copyright (C) 2010 Oohara Yuuma <oohara@libra.interq.or.jp>\n");
  printf("This program is free software;\n");
  printf("see the file LICENSE in the source code for details.\n");
  printf("last-of-candle comes with NO WARRANTY, to the extent "
         "permitted by law.\n");

  return 0;
}

static int
do_help(void)
{
  printf("Usage: last-of-candle [options]\n");
  printf("explore a dungeon with your shadows\n");
  printf("Options:\n");
  printf("--save-dir _dir_   save and load game in this directory\n");
  printf("--help             print this message\n");
  printf("--version          print version information\n");
  printf("\n");
  printf("Report bugs to <oohara@libra.interq.or.jp>.\n");

  return 0;
}

int
main(int argc, char *argv[])
{
  int i;
  int flag_help;
  int flag_version;
  world *wp = NULL;
  char *save_dir = NULL;

  flag_help = 0;
  flag_version = 0;
  for (i = 1; i < argc; i++)
  {
    if (!strcmp(argv[i], "--help"))
    {
      flag_help = 1;
    }
    else if (!strcmp(argv[i], "--version"))
    {
      flag_version = 1;
    }
    else if (!strcmp(argv[i], "--save-dir"))
    {
      if (i + 1 >= argc)
      {
        fprintf(stderr, "main: missing argument for --save-dir\n");
        return 1;
      }
      save_dir = argv[i + 1];
      i++;
    }
    else
    {
      fprintf(stderr, "main: unknown option: %s\n", argv[i]);
      fprintf(stderr, "Type %s --help for the list of options.\n", argv[0]);
      return 1;
    }
  }

  if (flag_version)
    do_version();
  if (flag_help)
    do_help();
  if ((flag_help) || (flag_version))
    return 0;

  if (check_window_size() != 0)
    return 1;

  srand((unsigned int) time(NULL));

  wp = world_new();
  if (wp == NULL)
  {
    fprintf(stderr, "main: world_new failed\n");
    return 1;
  }

  if (save_dir != NULL)
  {
    wp->save_dir = concat_string(1, save_dir);
  }
  if ((load_game(wp) != 0) || (!game_is_sane(wp, 1)))
  {
    if (build_new_game(wp) != 0)
    {
      fprintf(stderr, "main: build_new_game failed\n");
      return 1;
    }
    add_log(wp, "welcome");
  }
  else
  {
    add_log(wp, "welcome back");
  }

  if (world_init_curses(wp) != 0)
  {
    fprintf(stderr, "main: world_init_curses failed\n");
    return 1;
  }

  main_loop(wp);

  world_delete(wp);
  wp = NULL;

  return 0;
}
