/* $Id: monster.c,v 1.45 2010/03/12 03:10:39 oohara Exp $ */

#include <stdio.h>
/* rand, abs */
#include <stdlib.h>

#include "world.h"
#include "action.h"
#include "dungeon.h"
#include "creature-detail.h"
#include "array.h"

#include "monster.h"

static int action_decide_monster_do_nothing(world *wp, creature *who);
static int **player_distance_map(world *wp);
static int action_decide_monster_1(world *wp, creature *who);

static int
action_decide_monster_do_nothing(world *wp, creature *who)
{
  if (wp == NULL)
    return 1;
  if (who == NULL)
    return 1;

  who->act[0] = ACTION_WALK;
  who->act[1] = 0;
  who->act[2] = 0;

  return 0;
}

static int **
player_distance_map(world *wp)
{
  int i;
  int j;
  int dx;
  int dy;
  int x_wp;
  int y_wp;
  int x_result;
  int y_result;
  int dist;
  int queue_num;
  int queue_next_num;
  int already_in_queue_next;
  int **result = NULL;
  int **queue = NULL;
  int **queue_next = NULL;
  int **queue_temp = NULL;
  creature *player = NULL;

  if (wp == NULL)
    return NULL;
  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
    return NULL;
  if (wp->cr[wp->player_id] == NULL)
    return NULL;
  player = wp->cr[wp->player_id];

  result = array2_new(17, 17);
  if (result == NULL)
  {
    return NULL;
  }
  queue = array2_new(17 * 17, 2);
  if (queue == NULL)
  {
    array2_delete(result, 17, 17);
    result = NULL;
    return NULL;
  }
  queue_next = array2_new(17 * 17, 2);
  if (queue_next == NULL)
  {
    array2_delete(result, 17, 17);
    result = NULL;
    array2_delete(queue, 17 * 17, 2);
    queue = NULL;
    return NULL;
  }

  for (x_result = 0; x_result < 17; x_result++)
  {
    for (y_result = 0; y_result < 17; y_result++)
    {
      result[x_result][y_result] = -1;
    }
  }
  result[8][8] = 0;
  queue[0][0] = 8;
  queue[0][1] = 8;
  queue_num = 1;
  dist = 1;

  while (queue_num > 0)
  {
    queue_next_num = 0;
    for (i = 0; i < queue_num; i++)
    {
      for (dx = -1; dx <= 1; dx++)
      {
        for (dy = -1; dy <= 1; dy++)
        {
          if ((dx == 0) && (dy == 0))
            continue;
          x_result = queue[i][0] + dx;
          y_result = queue[i][1] + dy;
          if ((x_result < 0) || (x_result >= 17))
            continue;
          if ((y_result < 0) || (y_result >= 17))
            continue;
          if (result[x_result][y_result] >= 0)
            continue;
          x_wp = x_result - 8 + player->x;
          y_wp = y_result - 8 + player->y;
          if (grid_blocks_creature(wp, player->z, x_wp, y_wp))
            continue;
          result[x_result][y_result] = dist;
          already_in_queue_next = 0;
          for (j = 0; j < queue_next_num; j++)
          {
            if ((queue_next[j][0] == x_result)
                && (queue_next[j][1] == y_result))
            {
              already_in_queue_next = 1;
              break;
            }
          }
          if (!already_in_queue_next)
          { 
            queue_next[queue_next_num][0] = x_result;
            queue_next[queue_next_num][1] = y_result;
            queue_next_num++;
          }
        }
      }
    }
    queue_temp = queue;
    queue = queue_next;
    queue_next = queue_temp;
    queue_num = queue_next_num;
    dist++;
  }

  array2_delete(queue, 17 * 17, 2);
  queue = NULL;
  array2_delete(queue_next, 17 * 17, 2);
  queue_next = NULL;

  return result;
}

static int
action_decide_monster_1(world *wp, creature *who)
{
  int i;
  int dx;
  int dy;
  int x_distmap;
  int y_distmap;
  int dist_now;
  int move_num;
  struct creature_detail crd;
  creature *player = NULL;
  int **distmap = NULL;

  if (wp == NULL)
    return 1;
  if (who == NULL)
    return 1;

  creature_detail_get(who->type, &crd);

  if (who->act[0] != 0)
  {
    if (action_check(wp, who->id, who->act))
    {
      return 0;
    }
    else
    {
      for (i = 0; i < 4; i++)
        who->act[i] = 0;
    }
  }
  
  if (creature_is_dead(wp, wp->player_id))
  {
    action_decide_monster_do_nothing(wp, who);
    return 0;
  }
  player = wp->cr[wp->player_id];

  if (who->z != player->z)
  {
    action_decide_monster_do_nothing(wp, who);
    return 0;
  }

  switch (who->type)
  {
  case CR_PLAYER_SHADOW:
    who->act[0] = ACTION_STAY;
    return 0;
    break;
  case CR_DRAGON:
    if (who->hp >= crd.max_hp)
    {
      action_decide_monster_do_nothing(wp, who);
      return 0;
    }
    break;
  default:
    break;
  }

  who->act[0] = ACTION_WALK;
  who->act[1] = (rand() % 3) - 1;
  who->act[2] = (rand() % 3) - 1;

  if ((distance(who->x, who->y, player->x, player->y) > 8)
      || ((distance(who->x, who->y, player->x, player->y) > 6)
          && (rand() % 2 == 0)))
  {
    return 0;
  }

  distmap = player_distance_map(wp);

  x_distmap = (who->x - player->x) + 8;
  y_distmap = (who->y - player->y) + 8;
  dist_now = distmap[x_distmap][y_distmap];
  move_num = 0;

  for (dx = -1; dx <= 1; dx++)
  {
    for (dy = -1; dy <= 1; dy++)
    {
      if ((dx == 0) && (dy == 0))
        continue;
      if ((x_distmap + dx < 0) || (x_distmap + dx >= 17))
        continue;
      if ((y_distmap + dy < 0) || (y_distmap + dy >= 17))
        continue;
      if (distmap[x_distmap + dx][y_distmap + dy] < 0)
        continue;
      if (distmap[x_distmap + dx][y_distmap + dy] < dist_now)
      {
        if ((move_num <= 0) || (rand() % (move_num + 1) == 0))
        {  
          who->act[1] = dx;
          who->act[2] = dy;
        }
        move_num++;
      }
    }
  }

  array2_delete(distmap, 17, 17);
  distmap = NULL;

  return 0;
}

int
action_decide_monster(world *wp, int creature_id)
{
  creature *who = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "action_decide_monster: wp is NULL\n");
    return 1;
  }
  if ((creature_id < 0) || (creature_id >= wp->cr_size))
  {
    fprintf(stderr, "action_decide_monster: strange creature_id (%d)\n",
            creature_id);
    return 1;
  }
  if (creature_id == wp->player_id)
  {
    fprintf(stderr, "action_decide_monster: creature %d is player\n",
            creature_id);
    return 1;
  }

  if (creature_is_dead(wp, creature_id))
    return 0;
  who = wp->cr[creature_id];

  action_decide_monster_1(wp, who);
  
  if (!action_check(wp, who->id, who->act))
  {
    action_decide_monster_do_nothing(wp, who);
  }

  return 0;
}
