/* $Id: sanity.c,v 1.8 2010/03/27 15:49:37 oohara Exp $ */

#include <stdio.h>

#include "world.h"
#include "list-int.h"

#include "sanity.h"

int
game_is_sane(world *wp, int verbose)
{
  int i;
  int x;
  int y;
  int z;
  list_int *order = NULL;

  if (wp == NULL)
  {
    return 0;
  }

  if (wp->size_z <= 0)
  {
    return 0;
  }
  if (wp->size_x <= 0)
  {
    return 0;
  }
  if (wp->size_y <= 0)
  {
    return 0;
  }

  if (wp->grid == NULL)
  {
    return 0;
  }
  for (z = 0; z < wp->size_z; z++)
  {
    if (wp->grid[z] == NULL)
    {
      return 0;
    }
    for (x = 0; x < wp->size_x; x++)
    {
      if (wp->grid[z][x] == NULL)
      {
        return 0;
      }
      for (y = 0; y < wp->size_y; y++)
      {
        if (wp->grid[z][x][y] < 0)
        {
          return 0;
        }
      }
    }
  }

  if (wp->cr_size <= 0)
  {
    return 0;
  }
  if (wp->cr == NULL)
  {
    return 0;
  }
  for (i = 0; i < wp->cr_size; i++)
  {
    if (wp->cr[i] == NULL)
      continue;
    if (wp->cr[i]->id != i)
    {
      return 0;
    }
    if (wp->cr[i]->type < 0)
    {
      return 0;
    }
    if ((wp->cr[i]->z < 0) || (wp->cr[i]->z >= wp->size_z))
    {
      return 0;
    }
    if ((wp->cr[i]->x < 0) || (wp->cr[i]->x >= wp->size_x))
    {
      return 0;
    }
    if ((wp->cr[i]->y < 0) || (wp->cr[i]->x >= wp->size_y))
    {
      return 0;
    }
    if (wp->cr[i]->wait < 0)
    {
      return 0;
    }
    if (wp->cr[i]->act[0] < 0)
    {
      return 0;
    }
  }

  if (wp->obj_size <= 0)
  {
    return 0;
  }
  if (wp->obj == NULL)
  {
    return 0;
  }
  for (i = 0; i < wp->obj_size; i++)
  {
    if (wp->obj[i] == NULL)
      continue;
    if (wp->obj[i]->id != i)
    {
      return 0;
    }
    if (wp->obj[i]->type < 0)
    {
      return 0;
    }
    if ((wp->obj[i]->z < 0) || (wp->obj[i]->z >= wp->size_z))
    {
      return 0;
    }
    if ((wp->obj[i]->x < 0) || (wp->obj[i]->x >= wp->size_x))
    {
      return 0;
    }
    if ((wp->obj[i]->y < 0) || (wp->obj[i]->y >= wp->size_y))
    {
      return 0;
    }
  }

  if ((wp->turn_z < 0) || (wp->turn_z >= wp->size_z))
  {
    return 0;
  }
  if (wp->turn_order == NULL)
  {
    return 0;
  }
  for (i = 0; i < wp->size_z; i++)
  {
    if (wp->turn_order[i] == NULL)
      continue;

    order = wp->turn_order[i];
    if (order->prev != NULL)
    {
      return 0;
    }
    while (order != NULL)
    {
      if (wp->cr[order->n] == NULL)
      {
        return 0;
      }
      order = order->next;
    }
  }

  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
  {
    return 0;
  }

  if (wp->inv_size <= 0)
  {
    return 0;
  }
  if (wp->inv == NULL)
  {
    return 0;
  }
  for (i = 0; i < wp->inv_size; i++)
  {
    if (wp->inv[i] == NULL)
    {
      return 0;
    }
  }

  if (wp->label_size <= 0)
  {
    return 0;
  }
  if (wp->label == NULL)
  {
    return 0;
  }

  if (wp->mode == NULL)
  {
    return 0;
  }

  if (wp->log_size <= 0)
  {
    return 0;
  }
  if (wp->log == NULL)
  {
    return 0;
  }
  if ((wp->log_head < 0) || (wp->log_head >= wp->log_size))
  {
    return 0;
  }

  return 1;
}
