/* $Id: save.c,v 1.39 2010/03/12 13:14:04 oohara Exp $ */

/* FILE, rename */
#include <stdio.h>
/* free */
#include <stdlib.h>
/* stat */
#include <sys/stat.h>
/* errno */
#include <errno.h>

#include "world.h"
#include "int-loader.h"
#include "creature.h"
#include "object.h"
#include "list-int.h"
#include "inventory.h"
#include "dungeon.h"
#include "util.h"

#include "save.h"

#define SAVE_FILE_NAME "last-of-candle-save.txt"
#define OLD_SAVE_FILE_NAME "last-of-candle-save-old.txt"
#define TERMINATE_MARKER 0xdead
#define SAVE_FILE_VERSION 0

/* return 0 on success, 1 on error */
int
save_game(world *wp)
{
  int i;
  int j;
  int z;
  int x;
  int y;
  int error_found;
  char *save_path = NULL;
  FILE *fp = NULL;
  creature *who = NULL;
  list_int *order = NULL;
  object *what = NULL;
  inventory *inv = NULL;

  if (wp == NULL)
    return 1;

  if (wp->save_dir != NULL)
  {
    save_path = concat_string(3, wp->save_dir, "/", SAVE_FILE_NAME);
  }
  else
  {  
    save_path = concat_string(1, SAVE_FILE_NAME);
  }
  if (save_path == NULL)
    return 1;

  fp = fopen(save_path, "w");
  if (fp == NULL)
  {
    free(save_path);
    save_path = NULL;
    return 1;
  }

  error_found = 0;

  if (fprintf(fp, "# last-of-candle save file\n# DO NOT EDIT\n\n") < 0)
    error_found = 1;

  if (fprintf(fp, "# save file version\n%d\n\n", SAVE_FILE_VERSION) < 0)
    error_found = 1;

  if (fprintf(fp, "# creature\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n", wp->cr_size) < 0)
    error_found = 1;
  for (i = 0; i < wp->cr_size; i++)
  {
    who = wp->cr[i];
    if (fprintf(fp, "## %d\n", i) < 0)
      error_found = 1;
    if (who == NULL)
    {
      if (fprintf(fp, "-1\n") < 0)
        error_found = 1;
      continue;
    }
    if (fprintf(fp, "%d\n", who->id) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->type) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->z) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->x) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->y) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->wait) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->hp) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->enchant_type) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", who->enchant_duration) < 0)
      error_found = 1;
    for (j = 0; j < 4; j++)
    {
      if (fprintf(fp, "%d\n", who->act[j]) < 0)
        error_found = 1;
    }
  }
  if (fprintf(fp, "# creature end\n%#x\n\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "# object\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n", wp->obj_size) < 0)
    error_found = 1;
  for (i = 0; i < wp->obj_size; i++)
  {
    what = wp->obj[i];
    if (fprintf(fp, "## %d\n", i) < 0)
      error_found = 1;
    if (what == NULL)
    {
      if (fprintf(fp, "-1\n") < 0)
        error_found = 1;
      continue;
    }
    if (fprintf(fp, "%d\n", what->id) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", what->type) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", what->z) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", what->x) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", what->y) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", what->quantity) < 0)
      error_found = 1;
  }
  if (fprintf(fp, "# object end\n%#x\n\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "# grid\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n%d\n%d\n",
              wp->size_z, wp->size_x, wp->size_y) < 0)
    error_found = 1;

  for (z = 0; z < wp->size_z; z++)
  {
    for (x = 0; x < wp->size_x; x++)
    {
      if (fprintf(fp, "## (%d, %d, y)\n", z, x) < 0)
        error_found = 1;
      for (y = 0; y < wp->size_y; y++)
      {
        if (fprintf(fp, "%d\n", wp->grid[z][x][y]) < 0)
          error_found = 1;
      }
    }
  }
  if (fprintf(fp, "# grid end\n%#x\n\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "# turn_z\n%d\n", wp->turn_z) < 0)
    error_found = 1;

  if (fprintf(fp, "# turn_order\n") < 0)
    error_found = 1;
  for (i = 0; i < wp->size_z; i++)
  {  
    if (fprintf(fp, "## length %d\n%d\n\n", i,
                list_int_length_all(wp->turn_order[i])) < 0)
      error_found = 1;
    order = wp->turn_order[i];
    while (order != NULL)
    {
      if (fprintf(fp, "%d\n", order->n) < 0)
        error_found = 1;
      order = order->next;
    }
  }

  if (fprintf(fp, "# turn_order end\n%#x\n\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "# player_id\n%d\n", wp->player_id) < 0)
    error_found = 1;
  if (fprintf(fp, "# camera_z\n%d\n", wp->camera_z) < 0)
    error_found = 1;
  if (fprintf(fp, "# camera_x\n%d\n", wp->camera_x) < 0)
    error_found = 1;
  if (fprintf(fp, "# camera_y\n%d\n\n", wp->camera_y) < 0)
    error_found = 1;

  if (fprintf(fp, "# inventory\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n", wp->inv_size) < 0)
    error_found = 1;
  for (i = 0; i < wp->inv_size; i++)
  {
    if (fprintf(fp, "## %d\n", i) < 0)
      error_found = 1;
    if (wp->inv[i] == NULL)
    {
      if (fprintf(fp, "-1\n") < 0)
        error_found = 1;
      /* wp->inv[*] must be non-NULL */
      error_found = 1;
      continue;
    }
    inv = wp->inv[i];
    if (fprintf(fp, "%d\n", inv->id) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", inv->type) < 0)
      error_found = 1;
    if (fprintf(fp, "%d\n", inv->quantity) < 0)
      error_found = 1;
  }
  if (fprintf(fp, "# inventory end\n%#x\n\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "# label\n") < 0)
    error_found = 1;
  if (fprintf(fp, "## size\n%d\n\n", wp->label_size) < 0)
    error_found = 1;
  for (i = 0; i < wp->label_size; i++)
  {
    if (fprintf(fp, "%d\n", wp->label[i]) < 0)
      error_found = 1;
  }
  if (fprintf(fp, "# label end\n%#x\n\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fprintf(fp, "# victory\n%d\n", wp->victory) < 0)
    error_found = 1;
  if (fprintf(fp, "# shadow mode\n%d\n", wp->shadow_mode) < 0)
    error_found = 1;
  if (fprintf(fp, "# candle\n%d\n", wp->candle) < 0)
    error_found = 1;
  if (fprintf(fp, "# ut\n%d\n", wp->ut) < 0)
    error_found = 1;

  if (fprintf(fp, "\n# save file end\n%#x\n", TERMINATE_MARKER) < 0)
    error_found = 1;

  if (fclose(fp) != 0)
    error_found = 1;
  fp = NULL;

  free(save_path);
  save_path = NULL;

  if (error_found != 0)
    return 1;

  return 0;
}

/* return
 * 0 if a saved game is successfully loaded
 * 1 if no saved game is found
 * 2 on error
 */
int
load_game(world *wp)
{
  int i;
  int j;
  int z;
  int x;
  int y;
  int m;
  int n;
  int type;
  int quantity;
  int error_found;
  struct stat buf_stat;
  char *save_path = NULL;
  char *old_save_path = NULL;
  int_loader *ilp = NULL;
  creature *who = NULL;
  object *what = NULL;


  if (wp == NULL)
    return 2;

  world_reset_game(wp);

  if (wp->save_dir != NULL)
  {
    save_path = concat_string(3, wp->save_dir, "/", SAVE_FILE_NAME);
  }
  else
  {
    save_path = concat_string(1, SAVE_FILE_NAME);
  }

  if (save_path == NULL)
    return 2;

  errno = 0;
  if (stat(save_path, &buf_stat) != 0)
  {
    if (errno == ENOENT)
      return 1;
    return 2;
  }

  ilp = int_loader_new(save_path);
  if (ilp == NULL)
    return 2;

  /* save file version */
  if (int_loader_get(ilp) != SAVE_FILE_VERSION)
  {
    ilp->error_found = 1;
  }

  /* creature */
  m = int_loader_get(ilp);
  for (i = 0; i < m; i++)
  {
    n = int_loader_get(ilp);
    if (n < 0)
      continue;
    if (n != i)
    {
      ilp->error_found = 1;
      break;
    }
    who = creature_new();
    if (who == NULL)
    {
      ilp->error_found = 1;
      break;
    }
    who->id = n;
    who->type = int_loader_get(ilp);
    who->z = int_loader_get(ilp);
    who->x = int_loader_get(ilp);
    who->y = int_loader_get(ilp);
    who->wait = int_loader_get(ilp);
    who->hp = int_loader_get(ilp);
    who->enchant_type = int_loader_get(ilp);
    who->enchant_duration = int_loader_get(ilp);
    for (j = 0; j < 4; j++)
      who->act[j] = int_loader_get(ilp);
    if (add_creature_here(wp, who, n) < 0)
    {
      ilp->error_found = 1;
      creature_delete(who);
      who = NULL;
      break;
    }
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
    ilp->error_found = 1;
  }

  /* object */
  m = int_loader_get(ilp);
  for (i = 0; i < m; i++)
  {
    n = int_loader_get(ilp);
    if (n < 0)
      continue;
    if (n != i)
    {
      ilp->error_found = 1;
      break;
    }
    what = object_new();
    if (what == NULL)
    {
      ilp->error_found = 1;
      break;
    }
    what->id = n;
    what->type = int_loader_get(ilp);
    what->z = int_loader_get(ilp);
    what->x = int_loader_get(ilp);
    what->y = int_loader_get(ilp);
    what->quantity = int_loader_get(ilp);
    if (add_object_here(wp, what, n) < 0)
    {
      ilp->error_found = 1;
      object_delete(what);
      what = NULL;
      break;
    }
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
    ilp->error_found = 1;
  }

  /* grid */
  if (int_loader_get(ilp) != wp->size_z)
  {
    ilp->error_found = 1;
  }
  if (int_loader_get(ilp) != wp->size_x)
  {
    ilp->error_found = 1;
  }
  if (int_loader_get(ilp) != wp->size_y)
  {
    ilp->error_found = 1;
  }
  for (z = 0; z < wp->size_z; z++)
  {
    for (x = 0; x < wp->size_x; x++)
    {
      for (y = 0; y < wp->size_y; y++)
      {
        wp->grid[z][x][y] = int_loader_get(ilp);
      }
    }
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
    ilp->error_found = 1;
  }

  wp->turn_z = int_loader_get(ilp);

  /* turn order */
  for (i = 0; i < wp->size_z; i++)
  {  
    /* remove the list automatically created by add_creature_here() */
    if (wp->turn_order[i] != NULL)
    {
      list_int_delete_all(wp->turn_order[i]);
      wp->turn_order[i] = NULL;
    }
    m = int_loader_get(ilp);
    for (j = 0; j < m; j++)
      wp->turn_order[i] = list_int_append(wp->turn_order[i],
                                          int_loader_get(ilp));
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
    ilp->error_found = 1;
  }

  wp->player_id = int_loader_get(ilp);
  if ((wp->player_id < 0) || (wp->player_id >= wp->cr_size))
  {
    ilp->error_found = 1;
  }
  if (wp->cr[wp->player_id] == NULL)
  {
    ilp->error_found = 1;
  }
  wp->camera_z = int_loader_get(ilp);
  wp->camera_x = int_loader_get(ilp);
  wp->camera_y = int_loader_get(ilp);

  /* inventory */
  m = int_loader_get(ilp);
  for (i = 0; i < m; i++)
  {
    n = int_loader_get(ilp);
    if (n < 0)
    {
      ilp->error_found = 1;
      break;
    }
    type = int_loader_get(ilp);
    quantity = int_loader_get(ilp);

    if ((type < 0) || (quantity <= 0))
      continue;

    if (add_inventory_here(wp, type, quantity, n) < 0)
    {
      ilp->error_found = 1;
      break;
    }
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
    ilp->error_found = 1;
  }

  /* label */
  if (int_loader_get(ilp) != wp->label_size)
  {
    ilp->error_found = 1;
  }
  for (i = 0; i < wp->label_size; i++)
  {
    wp->label[i] = int_loader_get(ilp);
  }
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
    ilp->error_found = 1;
  }

  wp->victory = int_loader_get(ilp);
  wp->shadow_mode = int_loader_get(ilp);
  wp->candle = int_loader_get(ilp);
  wp->ut = int_loader_get(ilp);

  /* save file end */
  if (int_loader_get(ilp) != TERMINATE_MARKER)
  {
    ilp->error_found = 1;
  }

  if (ilp->error_found)
  {
    fprintf(stderr, "load error: ");
    if (ilp->error_message != NULL)
    {
      fprintf(stderr, "%s ", ilp->error_message);
    }
    fprintf(stderr, "at line %d\n", ilp->line_number);
  }

  error_found = ilp->error_found;
  int_loader_delete(ilp);
  ilp = NULL;

  /* the "save game" action in the previous game is done */
  who = wp->cr[wp->player_id];
  for (j = 0; j < 4; j++)
    who->act[j] = 0;

  wp->cursor_x = wp->camera_x;
  wp->cursor_y = wp->camera_y;

  if (wp->save_dir != NULL)
  {
    old_save_path = concat_string(3, wp->save_dir, "/", OLD_SAVE_FILE_NAME);
  }
  else
  {
    old_save_path = concat_string(1, OLD_SAVE_FILE_NAME);
  }

  if (old_save_path != NULL)
  {
    rename(save_path, old_save_path);
    free(old_save_path);
    old_save_path = NULL;
  }

  free(save_path);
  save_path = NULL;

  if (error_found != 0)
    return 2;

  return 0;
}
