/* $Id: world.c,v 1.83 2010/03/12 08:45:55 oohara Exp $ */

#include <stdio.h>
/* malloc, atexit */
#include <stdlib.h>

#include <curses.h>

#include "creature.h"
#include "list-int.h"
#include "array.h"
#include "object.h"
/* NUM_SCROLL */
#include "scroll-detail.h"

#include "world.h"

#define CANDLE_DEFAULT 6000

static void curses_cleanup(void);

static
void curses_cleanup(void)
{
  clear();
  refresh();
  nocbreak();
  echo();
  endwin();
}

world *
world_new(void)
{
  int i;
  int j;
  int k;
  world *wp = NULL;

  wp = (world *) malloc(sizeof(world));
  if (wp == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp) failed\n");
    return NULL;
  }

  wp->cr_size = 0;
  wp->cr = NULL;
  wp->obj_size = 0;
  wp->obj = NULL;
  wp->size_z = 0;
  wp->size_x = 0;
  wp->size_y = 0;
  wp->grid = NULL;
  wp->turn_z = -1;
  wp->turn_order = NULL;
  wp->player_id = -1;
  wp->camera_z = -1;
  wp->camera_x = -1;
  wp->camera_y = -1;
  wp->inv_size = 0;
  wp->inv = NULL;
  wp->label_size = 0;
  wp->label = NULL;
  wp->victory = 0;
  wp->shadow_mode = 0;
  wp->candle = CANDLE_DEFAULT;
  /* Unit of Time */
  wp->ut = 0;
  wp->scr = NULL;
  wp->save_dir = NULL;
  wp->should_quit = 0;
  wp->mode = NULL;
  wp->cursor_x = -1;
  wp->cursor_y = -1;
  wp->log_size = 0;
  wp->log = NULL;
  wp->log_head = 0;

  wp->cr_size = 4;
  wp->cr = (creature **) malloc(sizeof(creature *) * (wp->cr_size));
  if (wp->cr == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->cr) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->cr_size; i++)
    wp->cr[i] = NULL;

  wp->obj_size = 4;
  wp->obj = (object **) malloc(sizeof(object *) * wp->obj_size);
  if (wp->obj == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->obj) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->obj_size; i++)
    wp->obj[i] = NULL;

  wp->size_z = 11;
  wp->size_x = 50;
  wp->size_y = 50;
  wp->grid = array3_new(wp->size_z, wp->size_x, wp->size_y);
  if (wp->grid == NULL)
  {
    fprintf(stderr, "world_new: array3_new(wp->grid) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->size_z; i++)
  {
    for (j = 0; j < wp->size_x; j++)
    {
      for (k = 0; k < wp->size_y; k++)
      {
        wp->grid[i][j][k] = -1;
      }
    }
  }

  wp->turn_order = (list_int **) malloc(sizeof(list_int *) * wp->size_z);
  if (wp->turn_order == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->turn_order) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->size_z; i++)
    wp->turn_order[i] = NULL;

  wp->inv_size = 9;
  wp->inv = (inventory **) malloc(sizeof(inventory *) * wp->inv_size);
  if (wp->inv == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->inv) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->inv_size; i++)
    wp->inv[i] = NULL;
  for (i = 0; i < wp->inv_size; i++)
  {
    wp->inv[i] = inventory_new();
    if (wp->inv[i] == NULL)
    {
      fprintf(stderr, "world_new: inventory_new(%d) failed\n", i);
      world_delete(wp);
      wp = NULL;
      return NULL;
    }
    wp->inv[i]->id = i;
    wp->inv[i]->type = -1;
    wp->inv[i]->quantity = 0;
  }

  wp->label_size = NUM_SCROLL;
  wp->label = (int *) malloc(sizeof(int) * wp->label_size);
  if (wp->label == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->label) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < NUM_SCROLL; i++)
    wp->label[i] = i;

  wp->mode = (int *) malloc(sizeof(int) * 4);
  if (wp->mode == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->mode) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < 4; i++)
    wp->mode[i] = 0;

  wp->log_size = 50;
  wp->log = (char **) malloc(sizeof(char *) * wp->log_size);
  if (wp->log == NULL)
  {
    fprintf(stderr, "world_new: malloc(wp->log) failed\n");
    world_delete(wp);
    wp = NULL;
    return NULL;
  }
  for (i = 0; i < wp->log_size; i++)
    wp->log[i] = NULL;

  return wp;
}

void
world_delete(world *wp)
{
  int i;

  if (wp == NULL)
    return;

  if (wp->cr != NULL)
  {
    for (i = 0; i < wp->cr_size; i++)
    {
      if (wp->cr[i] != NULL)
      {
        creature_delete(wp->cr[i]);
        wp->cr[i] = NULL;
      }
    }
    free(wp->cr);
    wp->cr = NULL;
    wp->cr_size = 0;
  }
  
  if (wp->obj != NULL)
  {
    for (i = 0; i < wp->obj_size; i++)
    {
      if (wp->obj[i] != NULL)
      {
        object_delete(wp->obj[i]);
        wp->obj[i] = NULL;
      }
    }
    free(wp->obj);
    wp->obj = NULL;
    wp->obj_size = 0;
  }

  if (wp->grid != NULL)
  {
    array3_delete(wp->grid, wp->size_z, wp->size_x, wp->size_y);
    wp->grid = NULL;
    /* wp->size_z is also used for wp->turn_order --- don't reset here */
  }

  if (wp->turn_order != NULL)
  {
    for (i = 0; i < wp->size_z; i++)
    {
      if (wp->turn_order[i] != NULL)
      {
        list_int_delete_all(wp->turn_order[i]);
        wp->turn_order[i] = NULL;
      }
    }
    free(wp->turn_order);
    wp->turn_order = NULL;
  }
  wp->size_z = 0;
  wp->size_x = 0;
  wp->size_y = 0;

  if (wp->inv != NULL)
  {
    for (i = 0; i < wp->inv_size; i++)
    {  
      inventory_delete(wp->inv[i]);
      wp->inv[i] = NULL;
    }
    free(wp->inv);
    wp->inv = NULL;
    wp->inv_size = 0;
  }

  if (wp->label != NULL)
  {
    free(wp->label);
    wp->label = NULL;
    wp->label_size = 0;
  }

  if (wp->save_dir != NULL)
  {
    free(wp->save_dir);
    wp->save_dir = NULL;
  }

  if (wp->mode != NULL)
  {
    free(wp->mode);
    wp->mode = NULL;
  }

  if (wp->log != NULL)
  {
    for (i = 0; i < wp->log_size; i++)
    {
      if (wp->log[i] != NULL)
      {
        free(wp->log[i]);
        wp->log[i] = NULL;
      }
    }
    free(wp->log);
    wp->log = NULL;
    wp->log_size = 0;
  }

  free(wp);
  wp = NULL;
}

int
world_init_curses(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "world_init_curses: wp is NULL\n");
    return 1;
  }

  wp->scr = initscr();
  if (wp->scr == NULL)
  {
    fprintf(stderr, "world_init_curses: initscr failed\n");
    return 1;
  }
  if (atexit(&curses_cleanup) != 0)
  {
    fprintf(stderr, "world_new: atexit failed\n");
    curses_cleanup();
    return 1;
  }

  start_color();
  noecho();
  cbreak();
  keypad(wp->scr, TRUE);

  init_pair(1, COLOR_RED, COLOR_BLACK);
  init_pair(2, COLOR_GREEN, COLOR_BLACK);
  init_pair(3, COLOR_YELLOW, COLOR_BLACK);
  init_pair(4, COLOR_BLUE, COLOR_BLACK);
  init_pair(5, COLOR_MAGENTA, COLOR_BLACK);
  init_pair(6, COLOR_CYAN, COLOR_BLACK);
  init_pair(7, COLOR_WHITE, COLOR_BLACK);
  init_pair(8, COLOR_BLACK, COLOR_BLACK);

  return 0;
}

void
world_reset_game(world *wp)
{
  int i;
  int j;
  int k;

  if (wp == NULL)
  {
    fprintf(stderr, "world_reset_game: wp is NULL\n");
    return;
  }

  for (i = 0; i < wp->cr_size; i++)
  {
    if (wp->cr[i] != NULL)
    {
      creature_delete(wp->cr[i]);
      wp->cr[i] = NULL;
    }
  }

  for (i = 0; i < wp->obj_size; i++)
  {
    if (wp->obj[i] != NULL)
    {
      object_delete(wp->obj[i]);
      wp->obj[i] = NULL;
    }
  }

  for (i = 0; i < wp->size_z; i++)
  {
    for (j = 0; j < wp->size_x; j++)
    {
      for (k = 0; k < wp->size_y; k++)
      {
        wp->grid[i][j][k] = -1;
      }
    }
  }

  for (i = 0; i < wp->size_z; i++)
  {
    if (wp->turn_order[i] != NULL)
    {
      list_int_delete_all(wp->turn_order[i]);
      wp->turn_order[i] = NULL;
    }
  }

  wp->turn_z = -1;
  wp->player_id = -1;
  wp->camera_z = -1;
  wp->camera_x = -1;
  wp->camera_y = -1;

  for (i = 0; i < wp->inv_size; i++)
  {    
    wp->inv[i]->id = i;
    wp->inv[i]->type = -1;
    wp->inv[i]->quantity = 0;
  }

  for (i = 0; i < NUM_SCROLL; i++)
    wp->label[i] = i;

  wp->victory = 0;
  wp->shadow_mode = 0;
  wp->candle = CANDLE_DEFAULT;
  wp->ut = 0;
  wp->should_quit = 0;

  for (i = 0; i < 4; i++)
    wp->mode[i] = 0;

  wp->cursor_x = -1;
  wp->cursor_y = -1;

  for (i = 0; i < wp->log_size; i++)
  {
    if (wp->log[i] != NULL)
    {
      free(wp->log[i]);
      wp->log[i] = NULL;
    }
  }
  wp->log_head = 0;
}

int
world_expand_cr(world *wp)
{
  int i;
  int cr_size_temp;
  creature **cr_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "world_expand_cr: p is NULL\n");
    return 1;
  }

  cr_size_temp = wp->cr_size * 2;
  if (cr_size_temp <= 0)
  {
    fprintf(stderr, "world_expand_cr: cr_size overflowed\n");
    return 1;
  }
  cr_temp = (creature **) realloc(wp->cr, sizeof(creature *) * cr_size_temp);
  if (cr_temp == NULL)
  {
    fprintf(stderr, "world_expand_cr: realloc failed\n");
    return 1;
  }

  wp->cr = cr_temp;
  for (i = wp->cr_size; i < cr_size_temp; i++)
    wp->cr[i] = NULL;
  wp->cr_size = cr_size_temp;

  return 0;
}

int
world_expand_obj(world *wp)
{
  int i;
  int obj_size_temp;
  object **obj_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "world_expand_obj: wp is NULL\n");
    return 1;
  }

  obj_size_temp = wp->obj_size * 2;
  if (obj_size_temp <= 0)
  {
    fprintf(stderr, "world_expand_obj: obj_size overflowed\n");
    return 1;
  }
  obj_temp = (object **) realloc(wp->obj, sizeof(object *) * obj_size_temp);
  if (obj_temp == NULL)
  {
    fprintf(stderr, "world_expand_obj: realloc failed\n");
    return 1;
  }

  wp->obj = obj_temp;
  for (i = wp->obj_size; i < obj_size_temp; i++)
    wp->obj[i] = NULL;
  wp->obj_size = obj_size_temp;

  return 0;
}

int
world_expand_inv(world *wp)
{
  int i;
  int inv_size_temp;
  inventory **inv_temp = NULL;

  if (wp == NULL)
  {
    fprintf(stderr, "world_expand_inv: wp is NULL\n");
    return 1;
  }

  inv_size_temp = wp->inv_size * 2;
  if (inv_size_temp <= 0)
  {
    fprintf(stderr, "world_expand_inv: inv_size overflowed\n");
    return 1;
  }
  inv_temp = (inventory **) realloc(wp->inv,
                                    sizeof(inventory *) * inv_size_temp);
  if (inv_temp == NULL)
  {
    fprintf(stderr, "world_expand_inv: realloc failed\n");
    return 1;
  }

  wp->inv = inv_temp;
  for (i = wp->inv_size; i < inv_size_temp; i++)
    wp->inv[i] = NULL;
  for (i = wp->inv_size; i < inv_size_temp; i++)
  {  
    wp->inv[i] = inventory_new();
    if (wp->inv[i] == NULL)
    {
      fprintf(stderr, "world_expand_inv: inventory_new(%d) failed\n", i);
      return 1;
    }
    wp->inv[i]->id = i;
    wp->inv[i]->type = -1;
    wp->inv[i]->quantity = 0;
  }
  wp->inv_size = inv_size_temp;

  return 0;
}

